"""Specify how errors should be rendered."""
from enum import Enum

from astroid import nodes


def render_message(msg, node, source_lines):
    """Render a message based on type."""
    renderer = CUSTOM_MESSAGES.get(msg.symbol, render_generic)
    yield from renderer(msg, node, source_lines)


def render_generic(msg, node=None, source_lines=None):
    """Default rendering for a message."""
    if node is not None:
        start_line, start_col = node.fromlineno, node.col_offset

        if isinstance(node, (nodes.FunctionDef, nodes.ClassDef)):
            end_line, end_col = start_line, None
        else:
            end_line, end_col = node.end_lineno, node.end_col_offset

        # Display up to 2 lines before node for context:
        yield from render_context(start_line - 2, start_line, source_lines)

        if start_line == end_line:
            yield (
                start_line,
                slice(start_col, end_col),
                LineType.ERROR,
                source_lines[start_line - 1],
            )
        else:
            yield (start_line, slice(start_col, None), LineType.ERROR, source_lines[start_line - 1])
            yield from (
                (line, slice(None, None), LineType.ERROR, source_lines[line - 1])
                for line in range(start_line + 1, end_line)
            )
            yield (end_line, slice(None, end_col), LineType.ERROR, source_lines[end_line - 1])

        # Display up to 2 lines after node for context:
        yield from render_context(end_line + 1, end_line + 3, source_lines)

    else:
        line = msg.line
        yield from render_context(line - 2, line, source_lines)
        yield (line, slice(None, None), LineType.ERROR, source_lines[line - 1])
        yield from render_context(line + 1, line + 3, source_lines)


def render_missing_docstring(_msg, node, source_lines=None):
    """Render a missing docstring message."""
    if isinstance(node, nodes.Module):
        yield (None, slice(None, None), LineType.DOCSTRING, '"""YOUR DOCSTRING HERE"""')
        yield from render_context(1, 3, source_lines)
    elif isinstance(node, nodes.ClassDef) or isinstance(node, nodes.FunctionDef):
        start = node.fromlineno
        end = node.body[0].fromlineno
        yield from render_context(start, end, source_lines)
        # Calculate indentation
        body = source_lines[end - 1]
        indentation = len(body) - len(body.lstrip())
        yield (
            None,
            slice(None, None),
            LineType.DOCSTRING,
            body[:indentation] + '"""YOUR DOCSTRING HERE"""',
        )
        yield from render_context(end, end + 2, source_lines)


def render_trailing_newlines(msg, _node, source_lines=None):
    """Render a trailing newlines message."""
    start_line = msg.line - 1
    yield from render_context(start_line - 2, start_line, source_lines)
    yield from (
        (line, slice(None, None), LineType.OTHER, source_lines[line - 1])
        for line in range(start_line, len(source_lines) + 1)
    )


def render_context(start, stop, source_lines):
    """Helper for rendering context lines."""
    start, stop = max(start, 1), min(stop, len(source_lines))
    yield from (
        (line, slice(None, None), LineType.CONTEXT, source_lines[line - 1])
        for line in range(start, stop)
    )


def render_missing_return_type(_msg, node, source_lines=None):
    """Render a type annotation return message."""
    start_line, start_col = node.fromlineno, node.parent.col_offset
    end_line, end_col = node.end_lineno, node.end_col_offset

    # Display up to 2 lines before node for context:
    yield from render_context(start_line - 2, start_line, source_lines)
    yield from (
        (line, slice(None, end_col + 1), LineType.ERROR, source_lines[line - 1])
        for line in range(start_line, end_line + 1)
    )
    # Display up to 2 lines after node for context:
    yield from render_context(end_line + 1, end_line + 3, source_lines)


def render_too_many_arguments(msg, node, source_lines=None):
    """Render a too many arguments message."""
    # node is a FunctionDef node so replace it with its Arguments child
    yield from render_generic(msg, node.args, source_lines)


def render_missing_space_in_doctest(msg, _node, source_lines=None):
    """Render a missing space in doctest message"""
    line = msg.line

    # Display 2 lines before and after the erroneous line
    yield from render_context(line - 2, line, source_lines)
    yield (line, slice(None, None), LineType.ERROR, source_lines[line - 1])
    yield from render_context(line + 1, line + 3, source_lines)


CUSTOM_MESSAGES = {
    "missing-module-docstring": render_missing_docstring,
    "missing-class-docstring": render_missing_docstring,
    "missing-function-docstring": render_missing_docstring,
    "trailing-newlines": render_trailing_newlines,
    "missing-return-type": render_missing_return_type,
    "too-many-arguments": render_too_many_arguments,
    "missing-space-in-doctest": render_missing_space_in_doctest,
}


class LineType(Enum):
    """An enumeration for _add_line method line types."""

    ERROR = 1  # line with error
    CONTEXT = 2  # non-error/other line added for context
    OTHER = 3  # line included in source but not error
    ELLIPSIS = 5  # code replaced with ellipsis
    DOCSTRING = 6  # docstring needed warning



"""checker for type inference errors.
"""

from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

from python_ta.typecheck.base import TypeFail


class TypeInferenceChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "TypeInferenceChecker"
    msgs = {
        "E9900": (
            "Type error: %s",
            "type-error",
            "Presented when there is some kind of error with types.",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("type-error")
    def visit_default(self, node):
        if hasattr(node, "inf_type"):
            x = node.inf_type
            if isinstance(x, TypeFail):
                # If a child already has a type fail, this node inherited the
                # same one, so no need to report it again.
                for child in node.get_children():
                    if isinstance(getattr(child, "inf_type", None), TypeFail):
                        return

                # Otherwise, this is a new TypeFail that should be reported.
                self.add_message("type-error", args=x.msg, node=node)


def register(linter):
    linter.register_checker(TypeInferenceChecker(linter))



import inspect
import sys
import typing
from typing import Any, Callable, List, Optional, Set

import wrapt
from typeguard import check_type

DEBUG_CONTRACTS = False
"""
Set to True to display debugging messages when checking contracts.
"""

_DEFAULT_MAX_VALUE_LENGTH = 30


class PyTAContractError(Exception):
    """Error raised when a PyTA contract assertion is violated."""


def check_all_contracts(*mod_names: str, decorate_main: bool = True) -> None:
    """Automatically check contracts for all functions and classes in the given modules.

    By default (when called with no arguments), the current module is used.

    Args:
        *mod_names: The names of modules to check contracts for. These modules must have been
            previously imported.
        decorate_main: True if the module being run (where __name__ == '__main__') should
            have contracts checked.
    """

    modules = []
    if decorate_main:
        modules.append(sys.modules["__main__"])

    for module_name in mod_names:
        modules.append(sys.modules.get(module_name, None))

    for module in modules:
        if not module:
            # Module name was passed in incorrectly.
            continue
        for name, value in inspect.getmembers(module):
            if inspect.isfunction(value) or inspect.isclass(value):
                module.__dict__[name] = check_contracts(value)


@wrapt.decorator
def _enable_function_contracts(wrapped, instance, args, kwargs):
    """A decorator that enables checking contracts for a function."""
    try:
        if instance and inspect.isclass(instance):
            # This is a class method, so there is no instance.
            return _check_function_contracts(wrapped, None, args, kwargs)
        else:
            return _check_function_contracts(wrapped, instance, args, kwargs)
    except PyTAContractError as e:
        raise AssertionError(str(e)) from None


def check_contracts(func_or_class: Any) -> Callable:
    """A decorator to enable contract checking for a function or class.

    When used with a class, all methods defined within the class have contract checking enabled.

    Example:

        >>> from python_ta.contracts import check_contracts
        >>> @check_contracts
        ... def divide(x: int, y: int) -> int:
        ...     \"\"\"Return x // y.
        ...
        ...     Preconditions:
        ...        - y != 0
        ...     \"\"\"
        ...     return x // y
    """
    if inspect.isroutine(func_or_class):
        return _enable_function_contracts(func_or_class)
    elif inspect.isclass(func_or_class):
        add_class_invariants(func_or_class)
        return func_or_class
    else:
        # Default action
        return func_or_class


def add_class_invariants(klass: type) -> None:
    """Modify the given class to check representation invariants and method contracts."""
    if "__representation_invariants__" in klass.__dict__:
        # This means the class has already been decorated
        return

    # Update representation invariants from this class' docstring and those of its superclasses.
    rep_invariants = set()

    # Iterate over all inherited classes except builtins
    for cls in reversed(klass.__mro__):
        if "__representation_invariants__" in cls.__dict__:
            rep_invariants = rep_invariants.union(cls.__representation_invariants__)
        elif cls.__module__ != "builtins":
            rep_invariants.update(parse_assertions(cls, parse_token="Representation Invariant"))

    setattr(klass, "__representation_invariants__", rep_invariants)

    def new_setattr(self: klass, name: str, value: Any) -> None:
        """Set the value of the given attribute on self to the given value.

        Check representation invariants for this class when not within an instance method of the class.
        """
        cls_annotations = typing.get_type_hints(klass)

        if name in cls_annotations:
            try:
                _debug(f"Checking type of attribute {attr} for {klass.__qualname__} instance")
                check_type(name, value, cls_annotations[name])
            except TypeError:
                raise AssertionError(
                    f"Value {_display_value(value)} did not match type annotation for attribute "
                    f"{name}: {_display_annotation(cls_annotations[name])}"
                ) from None

        super(klass, self).__setattr__(name, value)
        curframe = inspect.currentframe()
        callframe = inspect.getouterframes(curframe, 2)
        frame_locals = callframe[1].frame.f_locals
        if self is not frame_locals.get("self"):
            # Only validating if the attribute is not being set in a instance/class method
            klass_mod = sys.modules.get(klass.__module__)
            if klass_mod is not None:
                try:
                    _check_invariants(self, klass, klass_mod.__dict__)
                except PyTAContractError as e:
                    raise AssertionError(str(e)) from None

    for attr, value in klass.__dict__.items():
        if inspect.isroutine(value):
            if isinstance(value, (staticmethod, classmethod)):
                # Don't check rep invariants for staticmethod and classmethod
                setattr(klass, attr, check_contracts(value))
            else:
                setattr(klass, attr, _instance_method_wrapper(value, klass))

    klass.__setattr__ = new_setattr


def _check_function_contracts(wrapped, instance, args, kwargs):
    params = wrapped.__code__.co_varnames[: wrapped.__code__.co_argcount]
    annotations = typing.get_type_hints(wrapped)
    args_with_self = args if instance is None else (instance,) + args

    # Check function parameter types
    for arg, param in zip(args_with_self, params):
        if param in annotations:
            try:
                _debug(f"Checking type of parameter {param} in call to {wrapped.__qualname__}")
                check_type(param, arg, annotations[param])
            except TypeError:
                additional_suggestions = _get_argument_suggestions(arg, annotations[param])

                raise PyTAContractError(
                    f"{wrapped.__name__} argument {_display_value(arg)} did not match type "
                    f"annotation for parameter {param}: {_display_annotation(annotations[param])}"
                    + (f"\n{additional_suggestions}" if additional_suggestions else "")
                )

    # Check function preconditions
    preconditions = parse_assertions(wrapped)
    function_locals = dict(zip(params, args_with_self))
    _check_assertions(wrapped, function_locals, preconditions)

    # Check return type
    r = wrapped(*args, **kwargs)
    if "return" in annotations:
        return_type = annotations["return"]
        try:
            _debug(f"Checking return type from call to {wrapped.__qualname__}")
            check_type("return", r, return_type)
        except TypeError:
            raise PyTAContractError(
                f"{wrapped.__name__}'s return value {_display_value(r)} did not match "
                f"return type annotation {_display_annotation(return_type)}"
            )

    return r


def _get_argument_suggestions(arg: Any, annotation: type) -> str:
    """Returns potential suggestions for the given arg and its annotation"""
    try:
        if isinstance(arg, type) and issubclass(arg, annotation):
            return "Did you mean {cls}(...) instead of {cls}?".format(cls=arg.__name__)
    except TypeError:
        pass

    return ""


def _instance_method_wrapper(wrapped: Callable, klass: type) -> Callable:
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        try:
            r = _check_function_contracts(wrapped, instance, args, kwargs)
            if _instance_init_in_callstack(instance):
                return r
            _check_class_type_annotations(klass, instance)
            klass_mod = sys.modules.get(klass.__module__)
            if klass_mod is not None:
                _check_invariants(instance, klass, klass_mod.__dict__)
        except PyTAContractError as e:
            raise AssertionError(str(e)) from None
        else:
            return r

    return wrapper(wrapped)


def _instance_init_in_callstack(instance: Any) -> bool:
    """Return whether instance's init is part of the current callstack

    Note: due to the nature of the check, externally defined __init__ functions with
    'self' defined as the first parameter may pass this check.
    """
    frame = inspect.currentframe().f_back
    while frame:
        frame_context_name = inspect.getframeinfo(frame).function
        frame_context_self = frame.f_locals.get("self")
        frame_context_vars = frame.f_code.co_varnames
        if (
            frame_context_name == "__init__"
            and frame_context_self is instance
            and frame_context_vars[0] == "self"
        ):
            return True
        frame = frame.f_back
    return False


def _check_class_type_annotations(klass: type, instance: Any) -> None:
    """Check that the type annotations for the class still hold.

    Precondition:
        - isinstance(instance, klass)
    """
    cls_annotations = typing.get_type_hints(klass)

    for attr, annotation in cls_annotations.items():
        value = getattr(instance, attr)
        try:
            _debug(f"Checking type of attribute {attr} for {klass.__qualname__} instance")
            check_type(attr, value, annotation)
        except TypeError:
            raise AssertionError(
                f"{_display_value(value)} did not match type annotation for attribute {attr}: "
                f"{_display_annotation(annotation)}"
            )


def _check_invariants(instance, klass: type, global_scope: dict) -> None:
    """Check that the representation invariants for the instance are satisfied."""
    rep_invariants = getattr(klass, "__representation_invariants__", set())

    for invariant in rep_invariants:
        try:
            _debug(
                "Checking representation invariant for "
                f"{instance.__class__.__qualname__}: {invariant}"
            )
            check = eval(invariant, {**global_scope, "self": instance})
        except:
            _debug(f"Warning: could not evaluate representation invariant: {invariant}")
        else:
            if not check:
                raise PyTAContractError(f'Representation invariant "{invariant}" was violated')


def _check_assertions(
    wrapped: Callable[..., Any], function_locals: dict, assertions: List[str]
) -> None:
    """Check that the given assertions are still satisfied."""
    for assertion in assertions:
        try:
            _debug(f"Checking precondition for {wrapped.__qualname__}: {assertion}")
            check = eval(assertion, {**wrapped.__globals__, **function_locals})
        except:
            _debug(f"Warning: could not evaluate precondition: {assertion}")
        else:
            if not check:
                arg_string = ", ".join(
                    f"{k}: {_display_value(v)}" for k, v in function_locals.items()
                )
                arg_string = "{" + arg_string + "}"
                raise PyTAContractError(
                    f'{wrapped.__name__} precondition "{assertion}" was '
                    f"violated for arguments {arg_string}"
                )


def parse_assertions(obj: Any, parse_token: str = "Precondition") -> List[str]:
    """Return a list of preconditions/representation invariants parsed from the given entity's docstring.

    Uses parse_token to determine what to look for. parse_token defaults to Precondition.

    Currently only supports two forms:

    1. A single line of the form "<parse_token>: <cond>"
    2. A group of lines starting with "<parse_token>s:", where each subsequent
       line is of the form "- <cond>". Each line is considered a separate condition.
       The lines can be separated by blank lines, but no other text.
    """
    docstring = getattr(obj, "__doc__") or ""
    lines = [line.strip() for line in docstring.split("\n")]
    assertion_lines = [
        i for i, line in enumerate(lines) if line.lower().startswith(parse_token.lower())
    ]

    if assertion_lines == []:
        return []

    first = assertion_lines[0]

    if lines[first].startswith(parse_token + ":"):
        return [lines[first][len(parse_token + ":") :].strip()]
    elif lines[first].startswith(parse_token + "s:"):
        assertions = []
        for line in lines[first + 1 :]:
            if line.startswith("-"):
                assertion = line[1:].strip()
                _debug(f"Adding assertion to {obj.__qualname__}: {assertion}")
                assertions.append(assertion)
            elif line != "":
                break
        return assertions
    else:
        return []


def _display_value(value: Any, max_length: int = _DEFAULT_MAX_VALUE_LENGTH) -> str:
    """Return a human-friendly representation of the given value.

    If DEBUG_CONTRACTS is False, truncate long strings to max_length characters.

    Preconditions:
        - max_length >= 5
    """
    s = repr(value)
    if not DEBUG_CONTRACTS and len(s) > max_length:
        i = (max_length - 3) // 2
        return s[:i] + "..." + s[-i:]
    else:
        return s


def _display_annotation(annotation: Any) -> str:
    """Return a human-friendly representation of the given type annotation.

    >>> _display_annotation(int)
    'int'
    >>> _display_annotation(list[int])
    'list[int]'
    >>> from typing import List
    >>> _display_annotation(List[int])
    'typing.List[int]'
    """
    if annotation is type(None):  # Use 'None' instead of 'NoneType'
        return "None"
    if hasattr(annotation, "__origin__"):  # Generic type annotations
        return repr(annotation)
    elif hasattr(annotation, "__name__"):
        return annotation.__name__
    else:
        return repr(annotation)


def _debug(msg: str) -> None:
    """Display a debugging message.

    Do nothing if DEBUG_CONTRACTS is False.
    """
    if not DEBUG_CONTRACTS:
        return

    print("[PyTA]", msg, file=sys.stderr)



from keyword import iskeyword
from typing import Callable, List, Tuple, Union

import astroid
import hypothesis.strategies as hs
from astroid import nodes
from hypothesis import assume, settings

from python_ta.transforms.type_inference_visitor import TypeInferer

settings.register_profile("pyta", settings(max_examples=10))


# Custom strategies for hypothesis testing framework
primitive_types = hs.sampled_from(
    [
        hs.integers,
        hs.booleans,
        lambda: hs.floats(allow_nan=False, allow_infinity=False),
        hs.none,
        hs.text,
        hs.binary,
    ]
)
primitive_values = primitive_types.flatmap(lambda s: s())


# Strategies for generating Indexes
index_types = hs.sampled_from(
    [hs.integers, lambda: hs.text(alphabet="abcdefghijklmnopqrstuvwxyz", min_size=1)]
)
index_values = index_types.flatmap(lambda s: s())


numeric_types = hs.sampled_from(
    [hs.integers, lambda: hs.floats(allow_nan=False, allow_infinity=False)]
)
numeric_values = numeric_types.flatmap(lambda s: s())


# Strategies for generating Binary Operators
non_bool_symbols = ["+", "-", "*", "//", "%", "/", "**", "&", "^", "|", "<<", ">>"]
non_boolean_operator = hs.sampled_from(non_bool_symbols)
non_bool_unary_op = hs.sampled_from(["-", "+", "~"])

# Strategy for genearting Comparison Operators
comparator_symbols = ["<", ">"]
comparator_operator = hs.sampled_from(comparator_symbols)
comparator_symbols_equality = ["==", "!=", ">=", "<=", "is"]
comparator_operator_equality = hs.sampled_from(comparator_symbols_equality)

# Strategy for generating Boolean Operators
binary_bool_operator = hs.sampled_from(["and", "or"])
unary_bool_operator = hs.sampled_from(["not"])

# Strategies for generating builtin type names
builtin_types = [
    bool,
    bytearray,
    bytes,
    complex,
    dict,
    enumerate,
    float,
    frozenset,
    int,
    list,
    set,
    str,
    tuple,
]
builtin_type = hs.sampled_from(builtin_types)
annotation = hs.sampled_from(builtin_types).map(lambda s: s.__name__)


def valid_identifier(**kwargs):
    """Return a strategy which generates a valid Python Identifier"""
    return hs.integers(min_value=0, max_value=1000).flatmap(lambda n: hs.just(f"x{n}"))


def homogeneous_list(**kwargs):
    """Return a strategy which generates a list of uniform type."""
    return primitive_types.flatmap(lambda s: hs.lists(s(), **kwargs))


def numeric_list(**kwargs):
    """Return a strategy which generates a list of uniform numeric types."""
    return numeric_types.flatmap(lambda s: hs.lists(s(), **kwargs))


def random_list(**kwargs):
    """Return a strategy which generates a random list."""
    return hs.lists(primitive_values, **kwargs)


def homogeneous_dictionary(**kwargs):
    """Return a strategy which generates a dictionary of uniform key:value type."""
    return index_types.flatmap(lambda s: hs.dictionaries(s(), s(), **kwargs))


def random_dictionary(**kwargs):
    """Return a strategy which generates a random list."""
    return hs.dictionaries(primitive_values, primitive_values, **kwargs)


def random_dict_variable_homogeneous_value(**kwargs):
    """Return a strategy which generates a random dictionary of variable name and value"""
    return primitive_types.flatmap(lambda s: hs.dictionaries(valid_identifier(), s(), **kwargs))


homogeneous_iterable = hs.sampled_from(
    [lambda: homogeneous_dictionary(min_size=1), lambda: homogeneous_list(min_size=1)]
).flatmap(lambda s: s())

heterogeneous_iterable = hs.sampled_from(
    [
        lambda: random_dictionary(min_size=1),
        lambda: random_list(min_size=1),
        lambda: hs.sets(primitive_values, min_size=1),
    ]
).flatmap(lambda s: s())


def _parse_dictionary_to_program(variables_dict):
    program = ""
    # parse dictionary into input program
    for variable_name in variables_dict:
        assume(not iskeyword(variable_name))
        program += variable_name + " = " + repr(variables_dict[variable_name]) + "\n"
    return program


# Strategies for generating Python ASTs.
# These are named after the corresponding nodes.
@hs.composite
def binop_node(draw, left=None, op=non_boolean_operator, right=None):
    left = left or const_node()
    right = right or const_node()
    node = nodes.BinOp(draw(op))
    node.postinit(draw(left), draw(right))
    return node


@hs.composite
def boolop_node(draw, value=None, op=binary_bool_operator, **kwargs):
    value = value or const_node()
    node = nodes.BoolOp(draw(op))
    if kwargs.get("min_size", 0) < 2:
        kwargs["min_size"] = 2
    node.postinit(draw(hs.lists(value, **kwargs)))
    return node


@hs.composite
def comprehension_node(draw, target=None, iter=None, ifs=hs.just([])):
    target = target or const_node(valid_identifier())
    iter = iter or list_node()
    node = nodes.Comprehension()
    node.postinit(draw(target), draw(iter), draw(ifs))
    return node


@hs.composite
def const_node(draw, value=primitive_values):
    """Return a Const node with value drawn from <value>."""
    return nodes.Const(draw(value))


@hs.composite
def dict_node(draw, key=const_node(), value=const_node(), **kwargs):
    items = draw(hs.dictionaries(key, value, **kwargs)).items()
    node = nodes.Dict()
    node.postinit(items)
    return node


@hs.composite
def expr_node(draw, value=None):
    value = value or expr
    node = nodes.Expr()
    node.postinit(draw(value))
    return node


@hs.composite
def ifexp_node(draw, test=const_node(hs.booleans()), expr=const_node(), orelse=const_node()):
    # TODO: Add an option for whether expr and orelse strategies produce the same type.
    test = draw(test)
    expr = draw(expr)
    node = nodes.IfExp()
    node.postinit(test, expr, expr)
    return node


@hs.composite
def index_node(draw, value=const_node(hs.integers())):
    node = nodes.Index()
    node.postinit(draw(value))
    return node


@hs.composite
def set_node(draw, elt=const_node(), **kwargs):
    """Return a Set node with elements drawn from elt."""
    node = nodes.Set()
    node.postinit(draw(hs.sets(elt, **kwargs)))
    return node


@hs.composite
def setcomp_node(draw, elt=const_node(), generators=hs.lists(comprehension_node(), min_size=1)):
    node = nodes.SetComp()
    node.postinit(draw(elt), draw(generators))
    return node


@hs.composite
def list_node(draw, elt=const_node(), **kwargs):
    """Return a List node with elements drawn from elt."""
    node = nodes.List()
    node.postinit(draw(hs.lists(elt, **kwargs)))
    return node


@hs.composite
def listcomp_node(draw, elt=const_node(), generators=hs.lists(comprehension_node(), min_size=1)):
    node = nodes.ListComp()
    node.postinit(draw(elt), draw(generators))
    return node


@hs.composite
def slice_node(draw):
    lower = draw(hs.one_of(const_node(hs.integers()), hs.none()))
    upper = draw(hs.one_of(const_node(hs.integers()), hs.none()))
    step = draw(hs.one_of(const_node(hs.integers()), hs.none()))
    node = nodes.Slice()
    node.postinit(lower, upper, step)
    return node


@hs.composite
def subscript_node(draw, value=None, slice=const_node(hs.integers())):
    value = value or subscriptable_expr
    node = nodes.Subscript()
    node.postinit(draw(value), draw(slice))
    return node


@hs.composite
def tuple_node(draw, elt=const_node, **kwargs):
    """Return a Tuple node with elements drawn from elt."""
    elts = draw(hs.lists(elt(), **kwargs, min_size=1))
    node = nodes.Tuple()
    node.postinit(elts)
    return node


@hs.composite
def unaryop_node(draw, op=hs.one_of(non_bool_unary_op, unary_bool_operator), operand=const_node()):
    op = draw(op)
    operand = draw(operand)
    node = nodes.UnaryOp(op)
    node.postinit(operand)
    return node


@hs.composite
def simple_homogeneous_dict_node(draw, **kwargs):
    k = draw(primitive_types)
    v = draw(primitive_types)
    return draw(dict_node(const_node(k()), const_node(v()), **kwargs))


@hs.composite
def simple_homogeneous_list_node(draw, **kwargs):
    t = draw(primitive_types)
    return draw(list_node(const_node(t()), **kwargs))


@hs.composite
def simple_homogeneous_set_node(draw, **kwargs):
    t = draw(primitive_types)
    homogeneous_set = draw(set_node(const_node(t()), **kwargs))
    assume(homogeneous_set.elts != set())
    return homogeneous_set


@hs.composite
def name_node(draw, name=None):
    if not name:
        node = nodes.Name(draw(valid_identifier()))
    else:
        node = nodes.Name(draw(name))
    return node


@hs.composite
def arguments_node(draw, annotated=False):
    n = draw(hs.integers(min_value=1, max_value=5))
    args = draw(hs.lists(name_node(None), min_size=n, max_size=n))
    if annotated:
        annotations = draw(hs.lists(name_node(annotation), min_size=n, max_size=n))
    else:
        annotations = None
    node = nodes.Arguments()
    node.postinit(args, None, None, None, annotations)
    return node


@hs.composite
def functiondef_node(draw, name=None, annotated=False, returns=False):
    name = name or draw(valid_identifier())
    args = draw(arguments_node(annotated))
    body = []
    returns_node = nodes.Return()
    arg_node, arg_type_node = draw(hs.sampled_from(list(zip(args.args, args.annotations))))
    if returns:
        returns_node.postinit(arg_node)
    else:
        returns_node.postinit(const_node(None))
    body.append(returns_node)
    node = nodes.FunctionDef(name=name)
    node.parent = nodes.Module("Default", None)
    node.postinit(args, body, None, arg_type_node)
    return node


expr = hs.one_of(const_node(), dict_node(min_size=1), list_node(min_size=1), tuple_node())

subscriptable_expr = hs.one_of(
    const_node(hs.text()), dict_node(min_size=1), list_node(min_size=1), tuple_node()
)


# Helper functions for testing
def _parse_text(
    source: Union[str, nodes.NodeNG], reset: bool = False
) -> Tuple[nodes.Module, TypeInferer]:
    """Parse source code text and output an AST with type inference performed."""
    # TODO: apparently no literal syntax for empty set in Python3, also cannot do set()
    # TODO: Deal with special case later.
    # if isinstance(source, nodes.Set) and len(list(source.elts)) == 0:
    #     source = f'{set({})}'
    if not isinstance(source, str):  # It's an astroid node
        source = source.as_string()
    module = astroid.parse(source)
    type_inferer = TypeInferer()
    if reset:
        type_inferer.reset()
    type_inferer.environment_transformer().visit(module)
    type_inferer.type_inference_transformer().visit(module)
    return module, type_inferer


def _verify_type_setting(module, ast_class, expected_type):
    """Helper to verify nodes visited by type inference visitor of astroid class has been properly transformed."""
    result = [n.inf_type.getValue() for n in module.nodes_of_class(ast_class)]
    assert [expected_type] == result, f"{expected_type}, {result}"


def lookup_type(inferer: TypeInferer, node: nodes.NodeNG, name: str) -> type:
    """Given a variable name, return its concrete type in the closest scope relative to given node.
    Should be used only for testing purposes.
    """
    inf_type = inferer.lookup_inf_type(node, name)
    return inf_type.getValue()


def types_in_callable(inferer: TypeInferer, callable_function: Callable) -> Tuple[List[type], type]:
    """Return a tuple of types corresponding to the Callable function's arguments and return value, respectively.
    Used only for testing purposes.
    """
    arg_type_lst = [
        inferer.type_constraints.resolve(argument).getValue()
        for argument in callable_function.__args__
    ]
    return arg_type_lst[:-1], arg_type_lst[-1]



import sys
from typing import *
from typing import ForwardRef

from astroid import nodes
from graphviz import Graph

from python_ta.transforms.type_inference_visitor import TypeInferer
from python_ta.typecheck.base import TypeFail, _TNode
from tests.custom_hypothesis_support import _parse_text

USAGE = "Usage: python -m sample_usage.draw_tnodes <your-file.py>"


def _type_str(type):
    if (
        isinstance(type, TypeVar)
        or isinstance(type, _GenericAlias)
        or type.__class__.__name__ == "_Any"
        or isinstance(type, ForwardRef)
        or type is None
    ):
        return str(type).replace("typing.", "")
    elif getattr(type, "__origin__", None) is Union:
        trimmed_args = []
        for arg in type.__args__:
            if not isinstance(arg, _GenericAlias):
                trimmed_args.append(_type_str(arg))
            else:
                break
        if len(trimmed_args) == 0:
            trimmed_args.append(_type_str(type.__args__[0]))
        if len(trimmed_args) != len(type.__args__):
            trimmed_args.append("...")
        return "Union[%s]" % ", ".join(trimmed_args)
    else:
        return type.__name__


def _find_type_fail(ast_node):
    if hasattr(ast_node, "inf_type") and isinstance(ast_node.inf_type, TypeFail):
        return ast_node.inf_type
    else:
        for child in ast_node.get_children():
            child_res = _find_type_fail(child)
            if child_res is not None:
                return child_res
    return None


def gen_graph_from_nodes(nodes, type_fail=None):
    graph = Graph(format="png", strict=True)
    graph.node_attr["fontname"] = "Courier New"
    graph.edge_attr["fontname"] = "Courier New"
    for node in nodes:
        graph.node(
            _type_str(node.type),
            "{type: %s|ast_node: %s|parent's type: %s}"
            % (
                _type_str(node.type),
                node.ast_node.as_string().replace("<", "\\<").replace(">", "\\>")
                if node.ast_node
                else "None",
                _type_str(node.parent.type) if node.parent else "NA",
            ),
            shape="record",
            style="rounded",
        )
        for neighb, ctx_node in node.adj_list:
            graph.edge(
                _type_str(node.type),
                _type_str(neighb.type),
                label=(f" {ctx_node.as_string()}" if ctx_node else ""),
            )
    if type_fail:
        graph.node(
            "tf",
            "{TypeFail|src_node: %s}"
            % (
                type_fail.src_node.as_string().replace("<", "\\<").replace(">", "\\>")
                if type_fail.src_node
                else "None"
            ),
            shape="record",
        )
        graph.edge("tf", _type_str(type_fail.tnode1.type), style="dashed")
        graph.edge("tf", _type_str(type_fail.tnode2.type), style="dashed")
    graph.view("tnode_graph")


def gen_graph_from_source(source: Union[str, nodes.NodeNG]):
    module, inferer = _parse_text(source)
    gen_graph_from_nodes(inferer.type_constraints._nodes, _find_type_fail(module))


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(USAGE)
        exit()

    with open(sys.argv[1]) as file:
        source = file.read()
        gen_graph_from_source(source)



import hashlib
import json
import uuid
from typing import Dict, List, NamedTuple

import requests


def errors_to_dict(errors: List[NamedTuple]) -> Dict[str, List[str]]:
    """Convert PyTA errors from MessageSet format to a json format Dictionary."""
    error_info = ["msg_id", "msg", "symbol", "module", "category", "line"]
    error_types = ["code", "style"]
    err_as_dict = {}
    for msg_set in errors:  # This iterates over the (filename, code, style) MessageSets
        for error_type in error_types:  # This iterates over the code and style attributes
            current_type = getattr(msg_set, error_type)  # Gets either the code or style dictionary
            for key in current_type.keys():  # Iterates over the error id's of caught errors
                err_as_dict[key] = []
                info_set = current_type.get(key)
                for (
                    msg
                ) in (
                    info_set.messages
                ):  # Iterates over the messages for each error of the given code
                    err_as_dict[key].append({k: getattr(msg, k) for k in error_info})
    return err_as_dict


def upload_to_server(
    errors: List[NamedTuple], paths: List[str], config: Dict[str, str], url: str, version: str
) -> None:
    """Send POST request to server with formatted data."""
    unique_id = get_hashed_id()  # Generates a device-specific ID
    files = []
    for path in paths:
        f = open(path)
        files.append(f)
    upload = {str(i): f for i, f in enumerate(files)}  # requests.post() requires passing a dict
    # 'upload' is an empty dict in the case that 'files' is empty
    errors_dict = errors_to_dict(errors)
    to_json = {"errors": errors_dict}
    if config:  # 'config' is an empty dictionary if the default was used
        to_json["cfg"] = config
    payload = json.dumps(to_json)
    try:
        response = requests.post(
            url=url, files=upload, data={"id": unique_id, "version": version, "payload": payload}
        )
        for f in files:
            f.close()
        response.raise_for_status()
        print("[INFO] Upload successful")
    except requests.HTTPError as e:
        print("[ERROR] Upload failed")
        if e.response.status_code == 400:
            print(
                "[ERROR] HTTP Response Status 400: Client-side error, likely due to improper syntax. "
                "Please report this to your instructor (and attach the code that caused the error)."
            )
        elif e.response.status_code == 403:
            print(
                "[ERROR] HTTP Response Status 403: Authorization is currently required for submission."
            )
        elif e.response.status_code == 500:
            print(
                "[ERROR] HTTP Response Status 500: The server ran into a situation it doesn't know how to handle. "
            )
            print(
                "Please report this to your instructor (and attach the code that caused the error)."
            )
        elif e.response.status_code == 503:
            print(
                "[ERROR] HTTP Response Status 503: The server is not ready to handle your request. "
            )
            print("It may be down for maintenance.")
        else:
            print('[ERROR] Error message: "{}"'.format(e))

    except requests.ConnectionError as e:
        print("[ERROR] Upload failed")
        print(
            "[ERROR] Error message: Connection timed out. This may be caused by your firewall, or the server may be "
            "temporarily down."
        )


def get_hashed_id() -> str:
    """
    Generates a unique ID by hashing the user's mac-address.
    """
    mac = str(uuid.uuid1())[24:]
    hash_gen = hashlib.sha512()
    encoded = mac.encode("utf-8")
    hash_gen.update(encoded)
    return hash_gen.hexdigest()



"""This module provides the core functionality for all PythonTA reporters.
"""
import os.path
import sys
from collections import defaultdict, namedtuple
from typing import Dict, List, Optional, Tuple

from astroid import NodeNG
from pylint.message import Message
from pylint.message.message_definition import MessageDefinition
from pylint.reporters import BaseReporter
from pylint.reporters.ureports.nodes import BaseLayout

from .node_printers import LineType, render_message

# Extension of Pylint's Message class to incorporate astroid node and source code snippet
NewMessage = namedtuple("NewMessage", Message._fields + ("node", "snippet"))

# Messages without a source code line to highlight
NO_SNIPPET = {"invalid-name"}


class PythonTaReporter(BaseReporter):
    """Abstract superclass for all PythonTA reporters.

    Reminder: see pylint BaseReporter for other instance variables.
    """

    # Rendering constants
    _SPACE = " "
    _BREAK = "\n"
    _COLOURING = {}
    _PRE_LINE_NUM_SPACES = 2

    # The error messages to report, mapping filename to a list of messages.
    messages: Dict[str, List[Message]]

    def __init__(self) -> None:
        """Initialize this reporter."""
        super().__init__()
        self.messages = defaultdict(list)
        self.source_lines = []
        self.module_name = ""
        self.current_file = ""

    def print_messages(self, level: str = "all") -> None:
        """Print messages for the current file.

        If level == 'all', both errors and style errors are displayed. Otherwise,
        only errors are displayed.
        """

    def has_messages(self) -> bool:
        """Return whether there are any messages registered."""
        return any(messages for messages in self.messages.values())

    def set_output(self, filepath: Optional[str] = None) -> None:
        """Set output stream based on filepath.

        If filepath is None or '-', sys.stdout is used.
        If filepath is the path to a file, that file is used (overwriting any existing contents).
        If filepath is the path to a directory, a new file is created in that directory
        (with default filename self.OUTPUT_FILENAME).
        """
        if filepath is None or filepath == "-":
            self.out = sys.stdout
        else:
            # Paths may contain system-specific or relative syntax, e.g. `~`, `../`
            filepath = os.path.expanduser(filepath)
            if os.path.isdir(filepath):
                filepath = os.path.join(filepath, self.OUTPUT_FILENAME)

            self.out = open(filepath, "w", encoding="utf-8")

    def handle_message(self, msg: Message) -> None:
        """Handle a new message triggered on the current file."""
        self.messages[self.current_file].append(msg)

    def handle_node(self, msg_definition: MessageDefinition, node: NodeNG) -> None:
        """Add node attribute to most recently-added message.

        This is used by our patched version of MessagesHandlerMixIn.add_message
        (see python_ta/patches/messages.py).
        """
        curr_messages = self.messages[self.current_file]
        if len(curr_messages) >= 1 and curr_messages[-1].msg_id == msg_definition.msgid:
            msg = curr_messages[-1]

            if msg.symbol in NO_SNIPPET or msg.msg.startswith("Invalid module"):
                snippet = ""
            else:
                snippet = self._build_snippet(msg, node)

            curr_messages[-1] = NewMessage(*msg, node, snippet)

    def group_messages(
        self, messages: List[Message]
    ) -> Tuple[Dict[str, List[Message]], Dict[str, List[Message]]]:
        """Group messages for the current file by their (error/style) and type (msg_id)."""
        error_msgs_by_type = defaultdict(list)
        style_msgs_by_type = defaultdict(list)
        for msg in messages:
            if msg.msg_id in ERROR_CHECKS or msg.symbol in ERROR_CHECKS:
                error_msgs_by_type[msg.msg_id].append(msg)
            else:
                style_msgs_by_type[msg.msg_id].append(msg)

        return error_msgs_by_type, style_msgs_by_type

    def display_messages(self, layout: BaseLayout) -> None:
        """Hook for displaying the messages of the reporter

        This will be called whenever the underlying messages
        needs to be displayed. For some reporters, it probably
        doesn't make sense to display messages as soon as they
        are available, so some mechanism of storing them could be used.
        This method can be implemented to display them after they've
        been aggregated.
        """

    # Rendering
    def _build_snippet(self, msg: Message, node: NodeNG) -> str:
        """Return a code snippet for the given Message object, formatted appropriately according
        to line type.
        """
        code_snippet = ""

        for lineno, slice_, line_type, text in render_message(msg, node, self.source_lines):
            code_snippet += self._add_line(lineno, line_type, slice_, text)

        return code_snippet

    def _add_line(self, lineno: int, linetype: LineType, slice_: slice, text: str = "") -> str:
        """Format given source code line as specified and return as str.

        Called by _build_snippet, relies on _colourify.
        """
        snippet = self._add_line_number(lineno, linetype)

        if linetype == LineType.ERROR:
            start_col = slice_.start or 0
            end_col = slice_.stop or len(text)

            if text[:start_col]:
                snippet += self._colourify("black", text[:start_col])
            snippet += self._colourify("highlight", text[slice_])
            if text[end_col:]:
                snippet += self._colourify("black", text[end_col:])
        elif linetype == LineType.CONTEXT:
            snippet += self._colourify("grey", text)
        elif linetype == LineType.OTHER:
            snippet += text
        elif linetype == LineType.DOCSTRING:
            space_c = len(text) - len(text.lstrip(" "))
            snippet += space_c * self._SPACE
            snippet += self._colourify("highlight", text.lstrip(" "))

        snippet += self._BREAK
        return snippet

    def _add_line_number(self, lineno: int, linetype: LineType) -> str:
        """Return a formatted string displaying a line number."""
        pre_spaces = self._PRE_LINE_NUM_SPACES * self._SPACE
        spaces = 2 * self._SPACE
        if lineno is not None:
            number = "{:>3}".format(lineno)
        else:
            number = 3 * self._SPACE

        if linetype == LineType.ERROR:
            return pre_spaces + self._colourify("gbold-line", number) + spaces
        elif linetype == LineType.CONTEXT:
            return pre_spaces + self._colourify("grey-line", number) + spaces
        elif linetype == LineType.OTHER:
            return pre_spaces + self._colourify("grey-line", number) + spaces
        elif linetype == LineType.DOCSTRING:
            return pre_spaces + self._colourify("black-line", number) + spaces
        else:
            return pre_spaces + number + spaces

    def _display(self, layout: BaseLayout) -> None:
        """display the layout"""

    @classmethod
    def _colourify(cls, colour_class: str, text: str) -> str:
        """Return a colourized version of text, using colour_class.

        By default, returns the text itself.
        """
        return text

    # Event callbacks
    def on_set_current_module(self, module: str, filepath: str) -> None:
        """Hook called when a module starts to be analysed."""
        self.module_name = module
        self.current_file = filepath

        if self.current_file not in self.messages:
            self.messages[self.current_file] = []

        with open(filepath, encoding="utf-8") as f:
            self.source_lines = [line.rstrip() for line in f.readlines()]

    def on_close(self, stats, previous_stats):
        """Hook called when a module finished analyzing.

        Close the reporter's output stream (if not sys.stdout).
        """
        if self.out is not sys.stdout:
            self.out.close()


# Checks to enable for basic_check (trying to find errors
# and forbidden constructs only)
ERROR_CHECKS = {
    "used-before-assignment",
    "undefined-variable",
    "undefined-loop-variable",
    "not-in-loop",
    "return-outside-function",
    "duplicate-key",
    "unreachable",
    "pointless-statement",
    "pointless-string-statement",
    "no-member",
    "not-callable",
    "assignment-from-no-return",
    "assignment-from-none",
    "no-value-for-parameter",
    "too-many-function-args",
    "invalid-sequence-index",
    "invalid-slice-index",
    "invalid-unary-operand-type",
    "unsupported-binary-operation",
    "unsupported-membership-test",
    "unsubscriptable-object",
    "unbalanced-tuple-unpacking",
    "unpacking-non-sequence",
    "function-redefined",
    "duplicate-argument-name",
    "import-error",
    "no-name-in-module",
    "non-parent-init-called",
    "access-member-before-definition",
    "method-hidden",
    "unexpected-special-method-signature",
    "inherit-non-class",
    "duplicate-except",
    "bad-except-order",
    "raising-bad-type",
    "raising-non-exception",
    "catching-non-exception",
    "bad-indentation",
    "E9996",
    "E9991",
    "E0001",
    "E9999",
    "unexpected-keyword-arg",
    "not-an-iterable",
    "nonexistent-operator",
    "invalid-length-returned",
    "abstract-method",
    "self-cls-assignment",
    "dict-iter-missing-items",
}



from typing import Set, _GenericAlias

from astroid import nodes
from hypothesis import HealthCheck, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.homogeneous_iterable)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_comprehension_reproduce_homogeneous(iterable):
    """Test SetComp node visitor representing a comprehension expression reproducing a set of elements of
    a homogeneous iterable."""
    program = "{elt for elt in " + repr(iterable) + "}"
    module, _ = cs._parse_text(program)
    setcomp_node = list(module.nodes_of_class(nodes.SetComp))[0]
    assert (
        setcomp_node.inf_type.getValue()
        == Set[setcomp_node.generators[0].iter.inf_type.getValue().__args__[0]]
    )


@given(cs.heterogeneous_iterable)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_comprehension_reproduce_heterogeneous(iterable):
    """Test SetComp node visitor representing a comprehension expression reproducing a set of elements of
    a heterogeneous iterable."""
    program = "{elt for elt in " + repr(iterable) + "}"
    module, _ = cs._parse_text(program)
    setcomp_node = list(module.nodes_of_class(nodes.SetComp))[0]
    assert (
        setcomp_node.inf_type.getValue()
        == Set[setcomp_node.generators[0].iter.inf_type.getValue().__args__[0]]
    )


@given(cs.valid_identifier(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_comprehension_reproduce_string(iterable):
    """Test SetComp node visitor representing a comprehension expression reproducing a set of elements of
    a string."""
    program = "{elt for elt in " + repr(iterable) + "}"
    module, _ = cs._parse_text(program)
    setcomp_node = list(module.nodes_of_class(nodes.SetComp))[0]
    assert (
        setcomp_node.inf_type.getValue() == Set[setcomp_node.generators[0].iter.inf_type.getValue()]
    )



"""checker for unnecessary indexing in a loop.
"""
from typing import List, Optional, Tuple, Union

import astroid
from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class UnnecessaryIndexingChecker(BaseChecker):
    __implements__ = IAstroidChecker
    # name is the same as file name but without _checker part
    name = "unnecessary_indexing"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9994": (
            'For loop variable "%s" can be simplified',
            "unnecessary-indexing",
            "Used when you have an loop variable in a for loop "
            "where its only usage is to index the iterable",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    # pass in message symbol as a parameter of check_messages
    @check_messages("unnecessary-indexing")
    def visit_for(self, node: nodes.For) -> None:
        if _is_unnecessary_indexing(node):
            args = node.target.name
            self.add_message("unnecessary-indexing", node=node.target, args=args)


# Helper functions
def _is_unnecessary_indexing(node: nodes.For) -> bool:
    """Return whether the iteration variable in the for loop is ONLY used to index the iterable.

    True if unnecessary usage, False otherwise or if iteration variable not used at all.
    """
    # Check if the iterable of the for loop is of the form "range(len(<variable-name>))".
    iterable = _iterable_if_range(node.iter)
    if iterable is None:
        return False

    index_nodes = _index_name_nodes(node.target.name, node)
    return all(_is_redundant(index_node, node) for index_node in index_nodes) and index_nodes


def _iterable_if_range(node: nodes.NodeNG) -> Optional[str]:
    """Return the iterable's name if this node is in "range" form, or None otherwise.

    Check for three forms:
      - range(len(<variable-name>))
      - range(0, len(<variable-name>))
      - range(0, len(<variable-name>), 1)
    """
    # Check outer function call is range
    if (
        not isinstance(node, nodes.Call)
        or not isinstance(node.func, nodes.Name)
        or not node.func.name == "range"
    ):
        return None

    # Check arguments to range
    if len(node.args) > 1:
        # Check that args[0] == Const(0)
        arg1 = node.args[0]
        if not isinstance(arg1, nodes.Const) or arg1.value != 0:
            return None
        if len(node.args) == 3 and (
            not isinstance(node.args[2], nodes.Const) or node.args[2].value != 1
        ):
            return None

    # Finally, check 'stop' argument is of the form len(<variable-name>).
    if len(node.args) == 1:
        stop_arg = node.args[0]
    else:
        stop_arg = node.args[1]

    if (
        isinstance(stop_arg, nodes.Call)
        and isinstance(stop_arg.func, nodes.Name)
        and stop_arg.func.name == "len"
        and len(stop_arg.args) == 1
        and isinstance(stop_arg.args[0], nodes.Name)
    ):
        return stop_arg.args[0].name


def _is_load_subscript(index_node: nodes.Name, for_node: nodes.For) -> bool:
    """Return whether or not <index_node> is used to subscript the iterable of <for_node>
    and the subscript item is being loaded from, e.g., s += iterable[index_node].

    NOTE: Index node is deprecated in Python 3.9

    Returns True if the following conditions are met:
    (3.9)
        - The <index_node> Name node is inside of a Subscript node
        - The item that is being indexed is the iterable of the for loop
        - The Subscript node is being used in a load context
    (3.8)
        - The <index_node> Name node is inside of an Index node
        - The Index node is inside of a Subscript node
        - The item that is being indexed is the iterable of the for loop
        - The Subscript node is being used in a load context
    """
    iterable = _iterable_if_range(for_node.iter)

    return (
        isinstance(index_node.parent, nodes.Subscript)
        and isinstance(index_node.parent.value, nodes.Name)
        and index_node.parent.value.name == iterable
        and index_node.parent.ctx == astroid.Load
    )


def _is_redundant(index_node: Union[nodes.AssignName, nodes.Name], for_node: nodes.For) -> bool:
    """Return whether or not <index_node> is redundant in <for_node>.

    The lookup method is used in case the original loop variable is shadowed
    in the for loop's body.
    """
    if isinstance(index_node, nodes.AssignName):
        previous_context = index_node.lookup(index_node.name)
        if previous_context[1] != ():
            return (
                previous_context[1][0] != for_node.target
                and previous_context[0] is for_node.scope()
            )
        else:
            return False
    else:
        return _scope_lookup(index_node) != for_node.target or _is_load_subscript(
            index_node, for_node
        )


def _index_name_nodes(index: str, for_node: nodes.For) -> List[Union[nodes.AssignName, nodes.Name]]:
    """Return a list of <index> AssignName and Name nodes contained in the body of <for_node>."""
    return [
        name_node
        for name_node in for_node.nodes_of_class((nodes.AssignName, nodes.Name))
        if name_node.name == index and name_node != for_node.target
    ]


def _scope_lookup(node: nodes.Name) -> Optional[nodes.NodeNG]:
    """Look up the given name node's assigment node.

    This is a replacement for astroid's LocalsDictNodeNG._scope_lookup method, which doesn't
    seem to handle nested comprehensions (?).
    """
    scope = node.scope()
    while node.name not in scope and not isinstance(scope, nodes.Module):
        scope = scope.parent.scope()

    if node.name in scope:
        return scope[node.name]
    else:
        return None


def register(linter):
    linter.register_checker(UnnecessaryIndexingChecker(linter))



from typing import Dict, List

from pylint.interfaces import IReporter

from .core import NewMessage, PythonTaReporter


class PlainReporter(PythonTaReporter):
    """Plain text reporter."""

    __implements__ = IReporter
    name = "PlainReporter"

    OUTPUT_FILENAME = "pyta_report.txt"

    # Rendering constants
    _SPACE = " "
    _BREAK = "\n"
    _COLOURING = {}
    code_err_title = "=== Code errors/forbidden usage (fix: high priority) ==="
    style_err_title = "=== Style/convention errors (fix: before submission) ==="
    no_err_message = "No problems detected, good job!" + _BREAK * 2
    no_snippet = "No code to display for this message." + _BREAK * 2

    def print_messages(self, level: str = "all") -> None:
        """Print messages for the current file.

        If level == 'all', both errors and style errors are displayed. Otherwise,
        only errors are displayed.
        """
        error_msgs, style_msgs = self.group_messages(self.messages[self.current_file])

        result = "PyTA Report for: " + self._colourify("bold", self.current_file) + self._BREAK
        result += self._colourify("code-heading", self.code_err_title + self._BREAK)
        messages_result = self._colour_messages_by_type(error_msgs)
        if messages_result:
            result += messages_result
        else:
            result += self.no_err_message

        if level == "all":
            result += self._colourify("style-heading", self.style_err_title + self._BREAK)
            messages_result = self._colour_messages_by_type(style_msgs)
            if messages_result:
                result += messages_result
            else:
                result += self.no_err_message

        self.writeln(result)

    def _colour_messages_by_type(self, messages: Dict[str, List[NewMessage]]) -> str:
        """
        Return string of properly formatted members of the messages dict
        (error or style) indicated by style.
        """
        max_messages = self.linter.config.pyta_number_of_messages

        result = ""
        for msg_id in messages:
            result += self._colourify("bold", msg_id)
            result += self._colourify("bold", " ({})  ".format(messages[msg_id][0].symbol))
            result += "Number of occurrences: {}.".format(len(messages[msg_id]))
            if max_messages != float("inf") and max_messages < len(messages[msg_id]):
                result += " (First {} shown).".format(max_messages)
            result += self._BREAK

            for i, msg in enumerate(messages[msg_id]):
                if i == max_messages:
                    break

                # Use only explanation, without redundant accessory information
                msg_truncated = msg.msg.split("\n")[0]
                result += 2 * self._SPACE
                result += (
                    self._colourify("bold", "[Line {}] {}".format(msg.line, msg_truncated))
                    + self._BREAK
                )

                result += msg.snippet
                result += self._BREAK

        return result



from typing import Tuple

from astroid import nodes
from hypothesis import HealthCheck, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.tuple_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_tuple(t_tuple):
    """Test Tuple nodes representing a tuple of various types."""
    module, _ = cs._parse_text(t_tuple)
    for t_node in module.nodes_of_class(nodes.Tuple):
        elt_types = tuple(elt.inf_type.getValue() for elt in t_node.elts)
        assert t_node.inf_type.getValue() == Tuple[elt_types]



"""Checker for E9973 missing-space-in-doctest"""
import re
from typing import Match, Optional, Union

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

DOCTEST = ">>>"


class MissingSpaceInDoctestChecker(BaseChecker):
    __implements__ = IAstroidChecker

    name = "missing_space_in_doctest"
    msgs = {
        "E9973": (
            'Space missing after >>> in the docstring of function "%s."',
            "missing-space-in-doctest",
            "Used when a doctest is missing a space before the code to be executed",
        )
    }
    # This is important so that your checker is executed before others
    priority = -1

    @check_messages("missing-space-in-doctest")
    def visit_functiondef(self, node: nodes.FunctionDef) -> None:
        """Visit a function definition"""
        docstring = node.doc

        if docstring is not None:
            start_line = node.lineno + 1
            lines = docstring.split("\n")

            for line_no, line in enumerate(lines):
                if self._has_invalid_doctest(line):
                    self.add_message(
                        "missing-space-in-doctest",
                        node=node,
                        args=node.name,
                        line=line_no + start_line,
                    )

    # Helper Function
    def _has_invalid_doctest(self, doc: str) -> Union[bool, Optional[Match[str]]]:
        """Return whether the docstring line contains an invalid doctest"""
        start_index = doc.find(DOCTEST)
        contains_doctest = start_index != -1
        if contains_doctest and len(doc) == 3:
            return True  # The doctest isn't followed by any character
        match = re.match("\s*>>>\w", doc)
        return match


def register(linter):
    """Required method to auto register this checker"""
    linter.register_checker(MissingSpaceInDoctestChecker(linter))



# True and False are deliberately omitted because they are keywords in
# Python 3, and stub files conform to Python 3 syntax.

import sys
from abc import ABCMeta, abstractmethod
from ast import AST, mod
from types import CodeType, TracebackType
from typing import (
    IO,
    AbstractSet,
    Any,
    AnyStr,
    BinaryIO,
    ByteString,
    Callable,
    Container,
    Dict,
    FrozenSet,
    Generic,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    NoReturn,
    Optional,
    Protocol,
    Reversible,
    Sequence,
    Set,
    Sized,
    SupportsAbs,
    SupportsComplex,
    SupportsFloat,
    SupportsInt,
    Text,
    Tuple,
    Type,
    TypeVar,
    Union,
    ValuesView,
    overload,
)

if sys.version_info >= (3,):
    from typing import SupportsBytes, SupportsRound

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_KT = TypeVar("_KT")
_VT = TypeVar("_VT")
_S = TypeVar("_S")
_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
_T3 = TypeVar("_T3")
_T4 = TypeVar("_T4")
_T5 = TypeVar("_T5")
_TT = TypeVar("_TT", bound="type")

class _SupportsIndex(Protocol):
    def __index__(self) -> int: ...

class object:
    __doc__: Optional[str]
    __dict__: Dict[str, Any]
    __slots__: Union[Text, Iterable[Text]]
    __module__: str
    if sys.version_info >= (3, 6):
        __annotations__: Dict[str, Any]
    @property
    def __class__(self: _T) -> Type[_T]: ...
    @__class__.setter
    def __class__(self, __type: Type[object]) -> None: ...
    def __init__(self) -> None: ...
    def __new__(cls) -> Any: ...
    def __setattr__(self, name: str, value: Any) -> None: ...
    def __eq__(self, o: object) -> bool: ...
    def __ne__(self, o: object) -> bool: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __hash__(self) -> int: ...
    def __format__(self, format_spec: str) -> str: ...
    def __getattribute__(self, name: str) -> Any: ...
    def __delattr__(self, name: str) -> None: ...
    def __sizeof__(self) -> int: ...
    def __reduce__(self) -> Union[str, Tuple[Any, ...]]: ...
    def __reduce_ex__(self, protocol: int) -> Union[str, Tuple[Any, ...]]: ...
    if sys.version_info >= (3,):
        def __dir__(self) -> Iterable[str]: ...
    if sys.version_info >= (3, 6):
        def __init_subclass__(cls) -> None: ...

class staticmethod(object):  # Special, only valid as a decorator.
    __func__: Callable[..., Any]
    if sys.version_info >= (3,):
        __isabstractmethod__: bool
    def __init__(self, f: Callable[..., Any]) -> None: ...
    def __new__(cls: Type[_T], *args: Any, **kwargs: Any) -> _T: ...
    def __get__(self, obj: _T, type: Optional[Type[_T]] = ...) -> Callable[..., Any]: ...

class classmethod(object):  # Special, only valid as a decorator.
    __func__: Callable[..., Any]
    if sys.version_info >= (3,):
        __isabstractmethod__: bool
    def __init__(self, f: Callable[..., Any]) -> None: ...
    def __new__(cls: Type[_T], *args: Any, **kwargs: Any) -> _T: ...
    def __get__(self, obj: _T, type: Optional[Type[_T]] = ...) -> Callable[..., Any]: ...

class type(object):
    __base__: type
    __bases__: Tuple[type, ...]
    __basicsize__: int
    __dict__: Dict[str, Any]
    __dictoffset__: int
    __flags__: int
    __itemsize__: int
    __module__: str
    __mro__: Tuple[type, ...]
    __name__: str
    if sys.version_info >= (3,):
        __qualname__: str
        __text_signature__: Optional[str]
    __weakrefoffset__: int
    @overload
    def __init__(self, o: object) -> None: ...
    @overload
    def __init__(self, name: str, bases: Tuple[type, ...], dict: Dict[str, Any]) -> None: ...
    @overload
    def __new__(cls, o: object) -> type: ...
    @overload
    def __new__(cls, name: str, bases: Tuple[type, ...], namespace: Dict[str, Any]) -> type: ...
    def __call__(self, *args: Any, **kwds: Any) -> Any: ...
    def __subclasses__(self: _TT) -> List[_TT]: ...
    # Note: the documentation doesnt specify what the return type is, the standard
    # implementation seems to be returning a list.
    def mro(self) -> List[type]: ...
    def __instancecheck__(self, instance: Any) -> bool: ...
    def __subclasscheck__(self, subclass: type) -> bool: ...
    if sys.version_info >= (3,):
        @classmethod
        def __prepare__(
            metacls, __name: str, __bases: Tuple[type, ...], **kwds: Any
        ) -> Mapping[str, Any]: ...

class super(object):
    if sys.version_info >= (3,):
        @overload
        def __init__(self, t: Any, obj: Any) -> None: ...
        @overload
        def __init__(self, t: Any) -> None: ...
        @overload
        def __init__(self) -> None: ...
    else:
        @overload
        def __init__(self, t: Any, obj: Any) -> None: ...
        @overload
        def __init__(self, t: Any) -> None: ...

class int:
    @overload
    def __init__(self, x: Union[Text, bytes, SupportsInt, _SupportsIndex] = ...) -> None: ...
    @overload
    def __init__(self, x: Union[Text, bytes, bytearray], base: int) -> None: ...
    if sys.version_info >= (3, 8):
        def as_integer_ratio(self) -> Tuple[int, Literal[1]]: ...
    @property
    def real(self) -> int: ...
    @property
    def imag(self) -> int: ...
    @property
    def numerator(self) -> int: ...
    @property
    def denominator(self) -> int: ...
    def conjugate(self) -> int: ...
    def bit_length(self) -> int: ...
    if sys.version_info >= (3,):
        def to_bytes(self, length: int, byteorder: str, *, signed: bool = ...) -> bytes: ...
        @classmethod
        def from_bytes(
            cls, bytes: Sequence[int], byteorder: str, *, signed: bool = ...
        ) -> int: ...  # TODO buffer object argument
    def __add__(self, x: int) -> int: ...
    def __sub__(self, x: int) -> int: ...
    def __mul__(self, x: int) -> int: ...
    def __floordiv__(self, x: int) -> int: ...
    if sys.version_info < (3,):
        def __div__(self, x: int) -> int: ...
    def __truediv__(self, x: int) -> float: ...
    def __mod__(self, x: int) -> int: ...
    def __divmod__(self, x: int) -> Tuple[int, int]: ...
    def __radd__(self, x: int) -> int: ...
    def __rsub__(self, x: int) -> int: ...
    def __rmul__(self, x: int) -> int: ...
    def __rfloordiv__(self, x: int) -> int: ...
    if sys.version_info < (3,):
        def __rdiv__(self, x: int) -> int: ...
    def __rtruediv__(self, x: int) -> float: ...
    def __rmod__(self, x: int) -> int: ...
    def __rdivmod__(self, x: int) -> Tuple[int, int]: ...
    def __pow__(
        self, __x: int, __modulo: Optional[int] = ...
    ) -> Any: ...  # Return type can be int or float, depending on x.
    def __rpow__(self, x: int) -> Any: ...
    def __and__(self, n: int) -> int: ...
    def __or__(self, n: int) -> int: ...
    def __xor__(self, n: int) -> int: ...
    def __lshift__(self, n: int) -> int: ...
    def __rshift__(self, n: int) -> int: ...
    def __rand__(self, n: int) -> int: ...
    def __ror__(self, n: int) -> int: ...
    def __rxor__(self, n: int) -> int: ...
    def __rlshift__(self, n: int) -> int: ...
    def __rrshift__(self, n: int) -> int: ...
    def __neg__(self) -> int: ...
    def __pos__(self) -> int: ...
    def __invert__(self) -> int: ...
    def __trunc__(self) -> int: ...
    if sys.version_info >= (3,):
        def __ceil__(self) -> int: ...
        def __floor__(self) -> int: ...
        def __round__(self, ndigits: Optional[int] = ...) -> int: ...
    def __getnewargs__(self) -> Tuple[int]: ...
    def __eq__(self, x: object) -> bool: ...
    def __ne__(self, x: object) -> bool: ...
    def __lt__(self, x: int) -> bool: ...
    def __le__(self, x: int) -> bool: ...
    def __gt__(self, x: int) -> bool: ...
    def __ge__(self, x: int) -> bool: ...
    def __str__(self) -> str: ...
    def __float__(self) -> float: ...
    def __int__(self) -> int: ...
    def __abs__(self) -> int: ...
    def __hash__(self) -> int: ...
    if sys.version_info >= (3,):
        def __bool__(self) -> bool: ...
    else:
        def __nonzero__(self) -> bool: ...
    def __index__(self) -> int: ...

class float:
    def __init__(
        self, x: Union[SupportsFloat, _SupportsIndex, Text, bytes, bytearray] = ...
    ) -> None: ...
    def as_integer_ratio(self) -> Tuple[int, int]: ...
    def hex(self) -> str: ...
    def is_integer(self) -> bool: ...
    @classmethod
    def fromhex(cls, s: str) -> float: ...
    @property
    def real(self) -> float: ...
    @property
    def imag(self) -> float: ...
    def conjugate(self) -> float: ...
    def __add__(self, x: float) -> float: ...
    def __sub__(self, x: float) -> float: ...
    def __mul__(self, x: float) -> float: ...
    def __floordiv__(self, x: float) -> float: ...
    if sys.version_info < (3,):
        def __div__(self, x: float) -> float: ...
    def __truediv__(self, x: float) -> float: ...
    def __mod__(self, x: float) -> float: ...
    def __divmod__(self, x: float) -> Tuple[float, float]: ...
    def __pow__(
        self, x: float
    ) -> float: ...  # In Python 3, returns complex if self is negative and x is not whole
    def __radd__(self, x: float) -> float: ...
    def __rsub__(self, x: float) -> float: ...
    def __rmul__(self, x: float) -> float: ...
    def __rfloordiv__(self, x: float) -> float: ...
    if sys.version_info < (3,):
        def __rdiv__(self, x: float) -> float: ...
    def __rtruediv__(self, x: float) -> float: ...
    def __rmod__(self, x: float) -> float: ...
    def __rdivmod__(self, x: float) -> Tuple[float, float]: ...
    def __rpow__(self, x: float) -> float: ...
    def __getnewargs__(self) -> Tuple[float]: ...
    def __trunc__(self) -> int: ...
    if sys.version_info >= (3,):
        @overload
        def __round__(self, ndigits: None = ...) -> int: ...
        @overload
        def __round__(self, ndigits: int) -> float: ...
    def __eq__(self, x: object) -> bool: ...
    def __ne__(self, x: object) -> bool: ...
    def __lt__(self, x: float) -> bool: ...
    def __le__(self, x: float) -> bool: ...
    def __gt__(self, x: float) -> bool: ...
    def __ge__(self, x: float) -> bool: ...
    def __neg__(self) -> float: ...
    def __pos__(self) -> float: ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    def __abs__(self) -> float: ...
    def __hash__(self) -> int: ...
    if sys.version_info >= (3,):
        def __bool__(self) -> bool: ...
    else:
        def __nonzero__(self) -> bool: ...

class complex:
    @overload
    def __init__(self, real: float = ..., imag: float = ...) -> None: ...
    @overload
    def __init__(self, real: Union[str, SupportsComplex, _SupportsIndex]) -> None: ...
    @property
    def real(self) -> float: ...
    @property
    def imag(self) -> float: ...
    def conjugate(self) -> complex: ...
    def __add__(self, x: complex) -> complex: ...
    def __sub__(self, x: complex) -> complex: ...
    def __mul__(self, x: complex) -> complex: ...
    def __pow__(self, x: complex) -> complex: ...
    if sys.version_info < (3,):
        def __div__(self, x: complex) -> complex: ...
    def __truediv__(self, x: complex) -> complex: ...
    def __radd__(self, x: complex) -> complex: ...
    def __rsub__(self, x: complex) -> complex: ...
    def __rmul__(self, x: complex) -> complex: ...
    def __rpow__(self, x: complex) -> complex: ...
    if sys.version_info < (3,):
        def __rdiv__(self, x: complex) -> complex: ...
    def __rtruediv__(self, x: complex) -> complex: ...
    def __eq__(self, x: object) -> bool: ...
    def __ne__(self, x: object) -> bool: ...
    def __neg__(self) -> complex: ...
    def __pos__(self) -> complex: ...
    def __str__(self) -> str: ...
    def __complex__(self) -> complex: ...
    def __abs__(self) -> float: ...
    def __hash__(self) -> int: ...
    if sys.version_info >= (3,):
        def __bool__(self) -> bool: ...
    else:
        def __nonzero__(self) -> bool: ...

if sys.version_info >= (3,):
    _str_base = object
else:
    class basestring(metaclass=ABCMeta): ...
    class unicode(basestring, Sequence[unicode]):
        @overload
        def __init__(self) -> None: ...
        @overload
        def __init__(self, o: object) -> None: ...
        @overload
        def __init__(self, o: str, encoding: unicode = ..., errors: unicode = ...) -> None: ...
        def capitalize(self) -> unicode: ...
        def center(self, width: int, fillchar: unicode = ...) -> unicode: ...
        def count(self, x: unicode) -> int: ...
        def decode(self, encoding: unicode = ..., errors: unicode = ...) -> unicode: ...
        def encode(self, encoding: unicode = ..., errors: unicode = ...) -> str: ...
        def endswith(
            self, suffix: Union[unicode, Tuple[unicode, ...]], start: int = ..., end: int = ...
        ) -> bool: ...
        def expandtabs(self, tabsize: int = ...) -> unicode: ...
        def find(self, sub: unicode, start: int = ..., end: int = ...) -> int: ...
        def format(self, *args: object, **kwargs: object) -> unicode: ...
        def index(self, sub: unicode, start: int = ..., end: int = ...) -> int: ...
        def isalnum(self) -> bool: ...
        def isalpha(self) -> bool: ...
        def isdecimal(self) -> bool: ...
        def isdigit(self) -> bool: ...
        def isidentifier(self) -> bool: ...
        def islower(self) -> bool: ...
        def isnumeric(self) -> bool: ...
        def isprintable(self) -> bool: ...
        def isspace(self) -> bool: ...
        def istitle(self) -> bool: ...
        def isupper(self) -> bool: ...
        def join(self, iterable: Iterable[unicode]) -> unicode: ...
        def ljust(self, width: int, fillchar: unicode = ...) -> unicode: ...
        def lower(self) -> unicode: ...
        def lstrip(self, chars: unicode = ...) -> unicode: ...
        def partition(self, sep: unicode) -> Tuple[unicode, unicode, unicode]: ...
        def replace(self, old: unicode, new: unicode, count: int = ...) -> unicode: ...
        def rfind(self, sub: unicode, start: int = ..., end: int = ...) -> int: ...
        def rindex(self, sub: unicode, start: int = ..., end: int = ...) -> int: ...
        def rjust(self, width: int, fillchar: unicode = ...) -> unicode: ...
        def rpartition(self, sep: unicode) -> Tuple[unicode, unicode, unicode]: ...
        def rsplit(self, sep: Optional[unicode] = ..., maxsplit: int = ...) -> List[unicode]: ...
        def rstrip(self, chars: unicode = ...) -> unicode: ...
        def split(self, sep: Optional[unicode] = ..., maxsplit: int = ...) -> List[unicode]: ...
        def splitlines(self, keepends: bool = ...) -> List[unicode]: ...
        def startswith(
            self, prefix: Union[unicode, Tuple[unicode, ...]], start: int = ..., end: int = ...
        ) -> bool: ...
        def strip(self, chars: unicode = ...) -> unicode: ...
        def swapcase(self) -> unicode: ...
        def title(self) -> unicode: ...
        def translate(self, table: Union[Dict[int, Any], unicode]) -> unicode: ...
        def upper(self) -> unicode: ...
        def zfill(self, width: int) -> unicode: ...
        @overload
        def __getitem__(self, i: int) -> unicode: ...
        @overload
        def __getitem__(self, s: slice) -> unicode: ...
        def __getslice__(self, start: int, stop: int) -> unicode: ...
        def __add__(self, s: unicode) -> unicode: ...
        def __mul__(self, n: int) -> unicode: ...
        def __rmul__(self, n: int) -> unicode: ...
        def __mod__(self, x: Any) -> unicode: ...
        def __eq__(self, x: object) -> bool: ...
        def __ne__(self, x: object) -> bool: ...
        def __lt__(self, x: unicode) -> bool: ...
        def __le__(self, x: unicode) -> bool: ...
        def __gt__(self, x: unicode) -> bool: ...
        def __ge__(self, x: unicode) -> bool: ...
        def __len__(self) -> int: ...
        # The argument type is incompatible with Sequence
        def __contains__(self, s: Union[unicode, bytes]) -> bool: ...  # type: ignore
        def __iter__(self) -> Iterator[unicode]: ...
        def __str__(self) -> str: ...
        def __repr__(self) -> str: ...
        def __int__(self) -> int: ...
        def __float__(self) -> float: ...
        def __hash__(self) -> int: ...
        def __getnewargs__(self) -> Tuple[unicode]: ...
    _str_base = basestring

class str(Sequence[str], _str_base):
    if sys.version_info >= (3,):
        @overload
        def __init__(self, o: object = ...) -> None: ...
        @overload
        def __init__(self, o: bytes, encoding: str = ..., errors: str = ...) -> None: ...
    else:
        def __init__(self, o: object = ...) -> None: ...
    def capitalize(self) -> str: ...
    if sys.version_info >= (3, 3):
        def casefold(self) -> str: ...
    def center(self, width: int, fillchar: str = ...) -> str: ...
    def count(self, x: Text, __start: Optional[int] = ..., __end: Optional[int] = ...) -> int: ...
    if sys.version_info < (3,):
        def decode(self, encoding: Text = ..., errors: Text = ...) -> unicode: ...
    def encode(self, encoding: Text = ..., errors: Text = ...) -> bytes: ...
    if sys.version_info >= (3,):
        def endswith(
            self,
            suffix: Union[Text, Tuple[Text, ...]],
            start: Optional[int] = ...,
            end: Optional[int] = ...,
        ) -> bool: ...
    else:
        def endswith(self, suffix: Union[Text, Tuple[Text, ...]]) -> bool: ...
    def expandtabs(self, tabsize: int = ...) -> str: ...
    def find(self, sub: Text, __start: Optional[int] = ..., __end: Optional[int] = ...) -> int: ...
    def format(self, *args: object, **kwargs: object) -> str: ...
    if sys.version_info >= (3,):
        def format_map(self, map: Mapping[str, Any]) -> str: ...
    def index(self, sub: Text, __start: Optional[int] = ..., __end: Optional[int] = ...) -> int: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    if sys.version_info >= (3, 7):
        def isascii(self) -> bool: ...
    if sys.version_info >= (3,):
        def isdecimal(self) -> bool: ...
    def isdigit(self) -> bool: ...
    if sys.version_info >= (3,):
        def isidentifier(self) -> bool: ...
    def islower(self) -> bool: ...
    if sys.version_info >= (3,):
        def isnumeric(self) -> bool: ...
        def isprintable(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    if sys.version_info >= (3,):
        def join(self, iterable: Iterable[str]) -> str: ...
    else:
        def join(self, iterable: Iterable[AnyStr]) -> AnyStr: ...
    def ljust(self, width: int, fillchar: str = ...) -> str: ...
    def lower(self) -> str: ...
    if sys.version_info >= (3,):
        def lstrip(self, chars: Optional[str] = ...) -> str: ...
        def partition(self, sep: str) -> Tuple[str, str, str]: ...
        def replace(self, old: str, new: str, count: int = ...) -> str: ...
    else:
        @overload
        def lstrip(self, chars: str = ...) -> str: ...
        @overload
        def lstrip(self, chars: unicode) -> unicode: ...
        @overload
        def partition(self, sep: bytearray) -> Tuple[str, bytearray, str]: ...
        @overload
        def partition(self, sep: str) -> Tuple[str, str, str]: ...
        @overload
        def partition(self, sep: unicode) -> Tuple[unicode, unicode, unicode]: ...
        def replace(self, old: AnyStr, new: AnyStr, count: int = ...) -> AnyStr: ...
    def rfind(self, sub: Text, __start: Optional[int] = ..., __end: Optional[int] = ...) -> int: ...
    def rindex(
        self, sub: Text, __start: Optional[int] = ..., __end: Optional[int] = ...
    ) -> int: ...
    def rjust(self, width: int, fillchar: str = ...) -> str: ...
    if sys.version_info >= (3,):
        def rpartition(self, sep: str) -> Tuple[str, str, str]: ...
        def rsplit(self, sep: Optional[str] = ..., maxsplit: int = ...) -> List[str]: ...
        def rstrip(self, chars: Optional[str] = ...) -> str: ...
        def split(self, sep: Optional[str] = ..., maxsplit: int = ...) -> List[str]: ...
    else:
        @overload
        def rpartition(self, sep: bytearray) -> Tuple[str, bytearray, str]: ...
        @overload
        def rpartition(self, sep: str) -> Tuple[str, str, str]: ...
        @overload
        def rpartition(self, sep: unicode) -> Tuple[unicode, unicode, unicode]: ...
        @overload
        def rsplit(self, sep: Optional[str] = ..., maxsplit: int = ...) -> List[str]: ...
        @overload
        def rsplit(self, sep: unicode, maxsplit: int = ...) -> List[unicode]: ...
        @overload
        def rstrip(self, chars: str = ...) -> str: ...
        @overload
        def rstrip(self, chars: unicode) -> unicode: ...
        @overload
        def split(self, sep: Optional[str] = ..., maxsplit: int = ...) -> List[str]: ...
        @overload
        def split(self, sep: unicode, maxsplit: int = ...) -> List[unicode]: ...
    def splitlines(self, keepends: bool = ...) -> List[str]: ...
    if sys.version_info >= (3,):
        def startswith(
            self,
            prefix: Union[Text, Tuple[Text, ...]],
            start: Optional[int] = ...,
            end: Optional[int] = ...,
        ) -> bool: ...
        def strip(self, chars: Optional[str] = ...) -> str: ...
    else:
        def startswith(self, prefix: Union[Text, Tuple[Text, ...]]) -> bool: ...
        @overload
        def strip(self, chars: str = ...) -> str: ...
        @overload
        def strip(self, chars: unicode) -> unicode: ...
    def swapcase(self) -> str: ...
    def title(self) -> str: ...
    if sys.version_info >= (3,):
        def translate(
            self, table: Union[Mapping[int, Union[int, str, None]], Sequence[Union[int, str, None]]]
        ) -> str: ...
    else:
        def translate(self, table: Optional[AnyStr], deletechars: AnyStr = ...) -> AnyStr: ...
    def upper(self) -> str: ...
    def zfill(self, width: int) -> str: ...
    if sys.version_info >= (3,):
        @staticmethod
        @overload
        def maketrans(
            x: Union[Dict[int, _T], Dict[str, _T], Dict[Union[str, int], _T]]
        ) -> Dict[int, _T]: ...
        @staticmethod
        @overload
        def maketrans(x: str, y: str, z: str = ...) -> Dict[int, Union[int, None]]: ...
    def __add__(self, s: str) -> str: ...
    # Incompatible with Sequence.__contains__
    def __contains__(self, o: Union[str, Text]) -> bool: ...  # type: ignore
    def __eq__(self, x: object) -> bool: ...
    def __ge__(self, x: Text) -> bool: ...
    def __getitem__(self, i: Union[int, slice]) -> str: ...
    def __gt__(self, x: Text) -> bool: ...
    def __hash__(self) -> int: ...
    def __iter__(self) -> Iterator[str]: ...
    def __le__(self, x: Text) -> bool: ...
    def __len__(self) -> int: ...
    def __lt__(self, x: Text) -> bool: ...
    def __mod__(self, x: Any) -> str: ...
    def __mul__(self, n: int) -> str: ...
    def __ne__(self, x: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __rmul__(self, n: int) -> str: ...
    def __str__(self) -> str: ...
    def __getnewargs__(self) -> Tuple[str]: ...
    if sys.version_info < (3,):
        def __getslice__(self, start: int, stop: int) -> str: ...
        def __float__(self) -> float: ...
        def __int__(self) -> int: ...

if sys.version_info >= (3,):
    class bytes(ByteString):
        @overload
        def __init__(self, ints: Iterable[int]) -> None: ...
        @overload
        def __init__(self, string: str, encoding: str, errors: str = ...) -> None: ...
        @overload
        def __init__(self, length: int) -> None: ...
        @overload
        def __init__(self) -> None: ...
        @overload
        def __init__(self, o: SupportsBytes) -> None: ...
        def capitalize(self) -> bytes: ...
        def center(self, width: int, fillchar: bytes = ...) -> bytes: ...
        def count(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def decode(self, encoding: str = ..., errors: str = ...) -> str: ...
        def endswith(self, suffix: Union[bytes, Tuple[bytes, ...]]) -> bool: ...
        def expandtabs(self, tabsize: int = ...) -> bytes: ...
        def find(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        if sys.version_info >= (3, 5):
            def hex(self) -> str: ...
        def index(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def isalnum(self) -> bool: ...
        def isalpha(self) -> bool: ...
        if sys.version_info >= (3, 7):
            def isascii(self) -> bool: ...
        def isdigit(self) -> bool: ...
        def islower(self) -> bool: ...
        def isspace(self) -> bool: ...
        def istitle(self) -> bool: ...
        def isupper(self) -> bool: ...
        def join(self, iterable: Iterable[Union[ByteString, memoryview]]) -> bytes: ...
        def ljust(self, width: int, fillchar: bytes = ...) -> bytes: ...
        def lower(self) -> bytes: ...
        def lstrip(self, chars: Optional[bytes] = ...) -> bytes: ...
        def partition(self, sep: bytes) -> Tuple[bytes, bytes, bytes]: ...
        def replace(self, old: bytes, new: bytes, count: int = ...) -> bytes: ...
        def rfind(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def rindex(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def rjust(self, width: int, fillchar: bytes = ...) -> bytes: ...
        def rpartition(self, sep: bytes) -> Tuple[bytes, bytes, bytes]: ...
        def rsplit(self, sep: Optional[bytes] = ..., maxsplit: int = ...) -> List[bytes]: ...
        def rstrip(self, chars: Optional[bytes] = ...) -> bytes: ...
        def split(self, sep: Optional[bytes] = ..., maxsplit: int = ...) -> List[bytes]: ...
        def splitlines(self, keepends: bool = ...) -> List[bytes]: ...
        def startswith(
            self,
            prefix: Union[bytes, Tuple[bytes, ...]],
            start: Optional[int] = ...,
            end: Optional[int] = ...,
        ) -> bool: ...
        def strip(self, chars: Optional[bytes] = ...) -> bytes: ...
        def swapcase(self) -> bytes: ...
        def title(self) -> bytes: ...
        def translate(self, table: Optional[bytes], delete: bytes = ...) -> bytes: ...
        def upper(self) -> bytes: ...
        def zfill(self, width: int) -> bytes: ...
        @classmethod
        def fromhex(cls, s: str) -> bytes: ...
        @classmethod
        def maketrans(cls, frm: bytes, to: bytes) -> bytes: ...
        def __len__(self) -> int: ...
        def __iter__(self) -> Iterator[int]: ...
        def __str__(self) -> str: ...
        def __repr__(self) -> str: ...
        def __int__(self) -> int: ...
        def __float__(self) -> float: ...
        def __hash__(self) -> int: ...
        @overload
        def __getitem__(self, i: int) -> int: ...
        @overload
        def __getitem__(self, s: slice) -> bytes: ...
        def __add__(self, s: bytes) -> bytes: ...
        def __mul__(self, n: int) -> bytes: ...
        def __rmul__(self, n: int) -> bytes: ...
        if sys.version_info >= (3, 5):
            def __mod__(self, value: Any) -> bytes: ...
        # Incompatible with Sequence.__contains__
        def __contains__(self, o: Union[int, bytes]) -> bool: ...  # type: ignore
        def __eq__(self, x: object) -> bool: ...
        def __ne__(self, x: object) -> bool: ...
        def __lt__(self, x: bytes) -> bool: ...
        def __le__(self, x: bytes) -> bool: ...
        def __gt__(self, x: bytes) -> bool: ...
        def __ge__(self, x: bytes) -> bool: ...
        def __getnewargs__(self) -> Tuple[bytes]: ...

else:
    bytes = str

class bytearray(MutableSequence[int], ByteString):
    if sys.version_info >= (3,):
        @overload
        def __init__(self) -> None: ...
        @overload
        def __init__(self, ints: Iterable[int]) -> None: ...
        @overload
        def __init__(self, string: Text, encoding: Text, errors: Text = ...) -> None: ...
        @overload
        def __init__(self, length: int) -> None: ...
    else:
        @overload
        def __init__(self) -> None: ...
        @overload
        def __init__(self, ints: Iterable[int]) -> None: ...
        @overload
        def __init__(self, string: str) -> None: ...
        @overload
        def __init__(self, string: Text, encoding: Text, errors: Text = ...) -> None: ...
        @overload
        def __init__(self, length: int) -> None: ...
    def capitalize(self) -> bytearray: ...
    def center(self, width: int, fillchar: bytes = ...) -> bytearray: ...
    if sys.version_info >= (3,):
        def count(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def copy(self) -> bytearray: ...
    else:
        def count(self, x: str) -> int: ...
    def decode(self, encoding: Text = ..., errors: Text = ...) -> str: ...
    def endswith(self, suffix: Union[bytes, Tuple[bytes, ...]]) -> bool: ...
    def expandtabs(self, tabsize: int = ...) -> bytearray: ...
    if sys.version_info < (3,):
        def extend(self, iterable: Union[str, Iterable[int]]) -> None: ...
    if sys.version_info >= (3,):
        def find(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        if sys.version_info >= (3, 5):
            def hex(self) -> str: ...
        def index(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
    else:
        def find(self, sub: str, start: int = ..., end: int = ...) -> int: ...
        def index(self, sub: str, start: int = ..., end: int = ...) -> int: ...
    def insert(self, index: int, object: int) -> None: ...
    def isalnum(self) -> bool: ...
    def isalpha(self) -> bool: ...
    if sys.version_info >= (3, 7):
        def isascii(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def islower(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    if sys.version_info >= (3,):
        def join(self, iterable: Iterable[Union[ByteString, memoryview]]) -> bytearray: ...
        def ljust(self, width: int, fillchar: bytes = ...) -> bytearray: ...
    else:
        def join(self, iterable: Iterable[str]) -> bytearray: ...
        def ljust(self, width: int, fillchar: str = ...) -> bytearray: ...
    def lower(self) -> bytearray: ...
    def lstrip(self, chars: Optional[bytes] = ...) -> bytearray: ...
    def partition(self, sep: bytes) -> Tuple[bytearray, bytearray, bytearray]: ...
    def replace(self, old: bytes, new: bytes, count: int = ...) -> bytearray: ...
    if sys.version_info >= (3,):
        def rfind(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
        def rindex(
            self, sub: Union[bytes, int], start: Optional[int] = ..., end: Optional[int] = ...
        ) -> int: ...
    else:
        def rfind(self, sub: bytes, start: int = ..., end: int = ...) -> int: ...
        def rindex(self, sub: bytes, start: int = ..., end: int = ...) -> int: ...
    def rjust(self, width: int, fillchar: bytes = ...) -> bytearray: ...
    def rpartition(self, sep: bytes) -> Tuple[bytearray, bytearray, bytearray]: ...
    def rsplit(self, sep: Optional[bytes] = ..., maxsplit: int = ...) -> List[bytearray]: ...
    def rstrip(self, chars: Optional[bytes] = ...) -> bytearray: ...
    def split(self, sep: Optional[bytes] = ..., maxsplit: int = ...) -> List[bytearray]: ...
    def splitlines(self, keepends: bool = ...) -> List[bytearray]: ...
    def startswith(
        self,
        prefix: Union[bytes, Tuple[bytes, ...]],
        start: Optional[int] = ...,
        end: Optional[int] = ...,
    ) -> bool: ...
    def strip(self, chars: Optional[bytes] = ...) -> bytearray: ...
    def swapcase(self) -> bytearray: ...
    def title(self) -> bytearray: ...
    if sys.version_info >= (3,):
        def translate(self, table: Optional[bytes], delete: bytes = ...) -> bytearray: ...
    else:
        def translate(self, table: str) -> bytearray: ...
    def upper(self) -> bytearray: ...
    def zfill(self, width: int) -> bytearray: ...
    @staticmethod
    def fromhex(s: str) -> bytearray: ...
    if sys.version_info >= (3,):
        @classmethod
        def maketrans(cls, frm: bytes, to: bytes) -> bytes: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[int]: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __int__(self) -> int: ...
    def __float__(self) -> float: ...
    __hash__: None  # type: ignore
    @overload
    def __getitem__(self, i: int) -> int: ...
    @overload
    def __getitem__(self, s: slice) -> bytearray: ...
    @overload
    def __setitem__(self, i: int, x: int) -> None: ...
    @overload
    def __setitem__(self, s: slice, x: Union[Iterable[int], bytes]) -> None: ...
    def __delitem__(self, i: Union[int, slice]) -> None: ...
    if sys.version_info < (3,):
        def __getslice__(self, start: int, stop: int) -> bytearray: ...
        def __setslice__(self, start: int, stop: int, x: Union[Sequence[int], str]) -> None: ...
        def __delslice__(self, start: int, stop: int) -> None: ...
    def __add__(self, s: bytes) -> bytearray: ...
    if sys.version_info >= (3,):
        def __iadd__(self, s: Iterable[int]) -> bytearray: ...
    def __mul__(self, n: int) -> bytearray: ...
    if sys.version_info >= (3,):
        def __rmul__(self, n: int) -> bytearray: ...
        def __imul__(self, n: int) -> bytearray: ...
    if sys.version_info >= (3, 5):
        def __mod__(self, value: Any) -> bytes: ...
    # Incompatible with Sequence.__contains__
    def __contains__(self, o: Union[int, bytes]) -> bool: ...  # type: ignore
    def __eq__(self, x: object) -> bool: ...
    def __ne__(self, x: object) -> bool: ...
    def __lt__(self, x: bytes) -> bool: ...
    def __le__(self, x: bytes) -> bool: ...
    def __gt__(self, x: bytes) -> bool: ...
    def __ge__(self, x: bytes) -> bool: ...

if sys.version_info >= (3,):
    _mv_container_type = int
else:
    _mv_container_type = str

class memoryview(Sized, Container[_mv_container_type]):
    format: str
    itemsize: int
    shape: Optional[Tuple[int, ...]]
    strides: Optional[Tuple[int, ...]]
    suboffsets: Optional[Tuple[int, ...]]
    readonly: bool
    ndim: int

    if sys.version_info >= (3,):
        c_contiguous: bool
        f_contiguous: bool
        contiguous: bool
        nbytes: int
        def __init__(self, obj: Union[bytes, bytearray, memoryview]) -> None: ...
        def __enter__(self) -> memoryview: ...
        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
        ) -> None: ...
    else:
        def __init__(self, obj: Union[bytes, bytearray, buffer, memoryview]) -> None: ...
    @overload
    def __getitem__(self, i: int) -> _mv_container_type: ...
    @overload
    def __getitem__(self, s: slice) -> memoryview: ...
    def __contains__(self, x: object) -> bool: ...
    def __iter__(self) -> Iterator[_mv_container_type]: ...
    def __len__(self) -> int: ...
    @overload
    def __setitem__(self, s: slice, o: memoryview) -> None: ...
    @overload
    def __setitem__(self, i: int, o: bytes) -> None: ...
    @overload
    def __setitem__(self, s: slice, o: Sequence[bytes]) -> None: ...
    def tobytes(self) -> bytes: ...
    def tolist(self) -> List[int]: ...
    if sys.version_info >= (3, 2):
        def release(self) -> None: ...
    if sys.version_info >= (3, 5):
        def hex(self) -> str: ...

class bool(int):
    def __init__(self, o: object = ...) -> None: ...
    @overload
    def __and__(self, x: bool) -> bool: ...
    @overload
    def __and__(self, x: int) -> int: ...
    @overload
    def __or__(self, x: bool) -> bool: ...
    @overload
    def __or__(self, x: int) -> int: ...
    @overload
    def __xor__(self, x: bool) -> bool: ...
    @overload
    def __xor__(self, x: int) -> int: ...
    @overload
    def __rand__(self, x: bool) -> bool: ...
    @overload
    def __rand__(self, x: int) -> int: ...
    @overload
    def __ror__(self, x: bool) -> bool: ...
    @overload
    def __ror__(self, x: int) -> int: ...
    @overload
    def __rxor__(self, x: bool) -> bool: ...
    @overload
    def __rxor__(self, x: int) -> int: ...
    def __getnewargs__(self) -> Tuple[int]: ...

class slice(object):
    start: Optional[int]
    step: Optional[int]
    stop: Optional[int]
    @overload
    def __init__(self, stop: Optional[int]) -> None: ...
    @overload
    def __init__(
        self, start: Optional[int], stop: Optional[int], step: Optional[int] = ...
    ) -> None: ...
    __hash__: None  # type: ignore
    def indices(self, len: int) -> Tuple[int, int, int]: ...

class tuple(Sequence[_T_co], Generic[_T_co]):
    def __new__(cls: Type[_T], iterable: Iterable[_T_co] = ...) -> _T: ...
    def __len__(self) -> int: ...
    def __contains__(self, x: object) -> bool: ...
    @overload
    def __getitem__(self, x: int) -> _T_co: ...
    @overload
    def __getitem__(self, x: slice) -> Tuple[_T_co, ...]: ...
    def __iter__(self) -> Iterator[_T_co]: ...
    def __lt__(self, x: Tuple[_T_co, ...]) -> bool: ...
    def __le__(self, x: Tuple[_T_co, ...]) -> bool: ...
    def __gt__(self, x: Tuple[_T_co, ...]) -> bool: ...
    def __ge__(self, x: Tuple[_T_co, ...]) -> bool: ...
    @overload
    def __add__(self, x: Tuple[_T_co, ...]) -> Tuple[_T_co, ...]: ...
    @overload
    def __add__(self, x: Tuple[Any, ...]) -> Tuple[Any, ...]: ...
    def __mul__(self, n: int) -> Tuple[_T_co, ...]: ...
    def __rmul__(self, n: int) -> Tuple[_T_co, ...]: ...
    def count(self, x: Any) -> int: ...
    if sys.version_info >= (3, 5):
        def index(self, x: Any, start: int = ..., end: int = ...) -> int: ...
    else:
        def index(self, x: Any) -> int: ...

# Added to handle zip calls (see tests/test_type_inference/test_for.py::test_for_zip)
class Iterator(Generic[_T]):
    def __iter__(self) -> Iterator[_T]: ...

class function:
    # TODO not defined in builtins!
    __name__: str
    __module__: str
    __code__: CodeType
    if sys.version_info >= (3,):
        __qualname__: str
        __annotations__: Dict[str, Any]

class list(MutableSequence[_T], Generic[_T]):
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[_T]) -> None: ...
    if sys.version_info >= (3,):
        def clear(self) -> None: ...
        def copy(self) -> List[_T]: ...
    def append(self, object: _T) -> None: ...
    def extend(self, iterable: Iterable[_T]) -> None: ...
    def pop(self, index: int = ...) -> _T: ...
    def index(self, object: _T, start: int = ..., stop: int = ...) -> int: ...
    def count(self, object: _T) -> int: ...
    def insert(self, index: int, object: _T) -> None: ...
    def remove(self, object: _T) -> None: ...
    def reverse(self) -> None: ...
    if sys.version_info >= (3,):
        def sort(
            self, *, key: Optional[Callable[[_T], Any]] = ..., reverse: bool = ...
        ) -> None: ...
    else:
        def sort(
            self,
            cmp: Callable[[_T, _T], Any] = ...,
            key: Callable[[_T], Any] = ...,
            reverse: bool = ...,
        ) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __str__(self) -> str: ...
    __hash__: None  # type: ignore
    @overload
    def __getitem__(self, i: int) -> _T: ...
    @overload
    def __getitem__(self, s: slice) -> List[_T]: ...
    @overload
    def __setitem__(self, i: int, o: _T) -> None: ...
    @overload
    def __setitem__(self, s: slice, o: Iterable[_T]) -> None: ...
    def __delitem__(self, i: Union[int, slice]) -> None: ...
    if sys.version_info < (3,):
        def __getslice__(self, start: int, stop: int) -> List[_T]: ...
        def __setslice__(self, start: int, stop: int, o: Sequence[_T]) -> None: ...
        def __delslice__(self, start: int, stop: int) -> None: ...
    def __add__(self, x: List[_T]) -> List[_T]: ...
    def __iadd__(self: _S, x: Iterable[_T]) -> _S: ...
    def __mul__(self, n: int) -> List[_T]: ...
    def __rmul__(self, n: int) -> List[_T]: ...
    if sys.version_info >= (3,):
        def __imul__(self: _S, n: int) -> _S: ...
    def __contains__(self, o: object) -> bool: ...
    def __reversed__(self) -> Iterator[_T]: ...
    def __gt__(self, x: List[_T]) -> bool: ...
    def __ge__(self, x: List[_T]) -> bool: ...
    def __lt__(self, x: List[_T]) -> bool: ...
    def __le__(self, x: List[_T]) -> bool: ...

class dict(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    # NOTE: Keyword arguments are special. If they are used, _KT must include
    #       str, but we have no way of enforcing it here.
    @overload
    def __init__(self, **kwargs: _VT) -> None: ...
    @overload
    def __init__(self, map: Mapping[_KT, _VT], **kwargs: _VT) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[Tuple[_KT, _VT]], **kwargs: _VT) -> None: ...
    def __new__(cls: Type[_T1], *args: Any, **kwargs: Any) -> _T1: ...
    if sys.version_info < (3,):
        def has_key(self, k: _KT) -> bool: ...
    def clear(self) -> None: ...
    def copy(self) -> Dict[_KT, _VT]: ...
    def popitem(self) -> Tuple[_KT, _VT]: ...
    def setdefault(self, k: _KT, default: _VT = ...) -> _VT: ...
    @overload
    def update(self, __m: Mapping[_KT, _VT], **kwargs: _VT) -> None: ...
    @overload
    def update(self, __m: Iterable[Tuple[_KT, _VT]], **kwargs: _VT) -> None: ...
    @overload
    def update(self, **kwargs: _VT) -> None: ...
    if sys.version_info >= (3,):
        def keys(self) -> KeysView[_KT]: ...
        def values(self) -> ValuesView[_VT]: ...
        def items(self) -> ItemsView[_KT, _VT]: ...
    else:
        def iterkeys(self) -> Iterator[_KT]: ...
        def itervalues(self) -> Iterator[_VT]: ...
        def iteritems(self) -> Iterator[Tuple[_KT, _VT]]: ...
        def viewkeys(self) -> KeysView[_KT]: ...
        def viewvalues(self) -> ValuesView[_VT]: ...
        def viewitems(self) -> ItemsView[_KT, _VT]: ...
    @staticmethod
    @overload
    def fromkeys(
        seq: Iterable[_T],
    ) -> Dict[_T, Any]: ...  # TODO: Actually a class method (mypy/issues#328)
    @staticmethod
    @overload
    def fromkeys(seq: Iterable[_T], value: _S) -> Dict[_T, _S]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, k: _KT) -> _VT: ...
    def __setitem__(self, k: _KT, v: _VT) -> None: ...
    def __delitem__(self, v: _KT) -> None: ...
    def __iter__(self) -> Iterator[_KT]: ...
    if sys.version_info >= (3, 8):
        def __reversed__(self) -> Iterator[_KT]: ...
    def __str__(self) -> str: ...
    __hash__: None  # type: ignore

class set(MutableSet[_T], Generic[_T]):
    def __init__(self, iterable: Iterable[_T] = ...) -> None: ...
    def add(self, element: _T) -> None: ...
    def clear(self) -> None: ...
    def copy(self) -> Set[_T]: ...
    def difference(self, *s: Iterable[Any]) -> Set[_T]: ...
    def difference_update(self, *s: Iterable[Any]) -> None: ...
    def discard(self, element: _T) -> None: ...
    def intersection(self, *s: Iterable[Any]) -> Set[_T]: ...
    def intersection_update(self, *s: Iterable[Any]) -> None: ...
    def isdisjoint(self, s: Iterable[Any]) -> bool: ...
    def issubset(self, s: Iterable[Any]) -> bool: ...
    def issuperset(self, s: Iterable[Any]) -> bool: ...
    def pop(self) -> _T: ...
    def remove(self, element: _T) -> None: ...
    def symmetric_difference(self, s: Iterable[_T]) -> Set[_T]: ...
    def symmetric_difference_update(self, s: Iterable[_T]) -> None: ...
    def union(self, *s: Iterable[_T]) -> Set[_T]: ...
    def update(self, *s: Iterable[_T]) -> None: ...
    def __len__(self) -> int: ...
    def __contains__(self, o: object) -> bool: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __str__(self) -> str: ...
    def __and__(self, s: AbstractSet[object]) -> Set[_T]: ...
    def __iand__(self, s: AbstractSet[object]) -> Set[_T]: ...
    def __or__(self, s: AbstractSet[_S]) -> Set[Union[_T, _S]]: ...
    def __ior__(self, s: AbstractSet[_S]) -> Set[Union[_T, _S]]: ...
    def __sub__(self, s: AbstractSet[object]) -> Set[_T]: ...
    def __isub__(self, s: AbstractSet[object]) -> Set[_T]: ...
    def __xor__(self, s: AbstractSet[_S]) -> Set[Union[_T, _S]]: ...
    def __ixor__(self, s: AbstractSet[_S]) -> Set[Union[_T, _S]]: ...
    def __le__(self, s: AbstractSet[object]) -> bool: ...
    def __lt__(self, s: AbstractSet[object]) -> bool: ...
    def __ge__(self, s: AbstractSet[object]) -> bool: ...
    def __gt__(self, s: AbstractSet[object]) -> bool: ...
    __hash__: None  # type: ignore

class frozenset(AbstractSet[_T], Generic[_T]):
    def __init__(self, iterable: Iterable[_T] = ...) -> None: ...
    def copy(self) -> FrozenSet[_T]: ...
    def difference(self, *s: Iterable[object]) -> FrozenSet[_T]: ...
    def intersection(self, *s: Iterable[object]) -> FrozenSet[_T]: ...
    def isdisjoint(self, s: Iterable[_T]) -> bool: ...
    def issubset(self, s: Iterable[object]) -> bool: ...
    def issuperset(self, s: Iterable[object]) -> bool: ...
    def symmetric_difference(self, s: Iterable[_T]) -> FrozenSet[_T]: ...
    def union(self, *s: Iterable[_T]) -> FrozenSet[_T]: ...
    def __len__(self) -> int: ...
    def __contains__(self, o: object) -> bool: ...
    def __iter__(self) -> Iterator[_T]: ...
    def __str__(self) -> str: ...
    def __and__(self, s: AbstractSet[_T]) -> FrozenSet[_T]: ...
    def __or__(self, s: AbstractSet[_S]) -> FrozenSet[Union[_T, _S]]: ...
    def __sub__(self, s: AbstractSet[_T]) -> FrozenSet[_T]: ...
    def __xor__(self, s: AbstractSet[_S]) -> FrozenSet[Union[_T, _S]]: ...
    def __le__(self, s: AbstractSet[object]) -> bool: ...
    def __lt__(self, s: AbstractSet[object]) -> bool: ...
    def __ge__(self, s: AbstractSet[object]) -> bool: ...
    def __gt__(self, s: AbstractSet[object]) -> bool: ...

class enumerate(Iterator[Tuple[int, _T]], Generic[_T]):
    def __init__(self, iterable: Iterable[_T], start: int = ...) -> None: ...
    def __iter__(self) -> Iterator[Tuple[int, _T]]: ...
    if sys.version_info >= (3,):
        def __next__(self) -> Tuple[int, _T]: ...
    else:
        def next(self) -> Tuple[int, _T]: ...

if sys.version_info >= (3,):
    class range(Sequence[int]):
        start: int
        stop: int
        step: int
        @overload
        def __init__(self, stop: int) -> None: ...
        @overload
        def __init__(self, start: int, stop: int, step: int = ...) -> None: ...
        def count(self, value: int) -> int: ...
        def index(self, value: int, start: int = ..., stop: Optional[int] = ...) -> int: ...
        def __len__(self) -> int: ...
        def __contains__(self, o: object) -> bool: ...
        def __iter__(self) -> Iterator[int]: ...
        @overload
        def __getitem__(self, i: int) -> int: ...
        @overload
        def __getitem__(self, s: slice) -> range: ...
        def __repr__(self) -> str: ...
        def __reversed__(self) -> Iterator[int]: ...

else:
    class xrange(Sized, Iterable[int], Reversible[int]):
        @overload
        def __init__(self, stop: int) -> None: ...
        @overload
        def __init__(self, start: int, stop: int, step: int = ...) -> None: ...
        def __len__(self) -> int: ...
        def __iter__(self) -> Iterator[int]: ...
        def __getitem__(self, i: int) -> int: ...
        def __reversed__(self) -> Iterator[int]: ...

class property(object):
    def __init__(
        self,
        fget: Optional[Callable[[Any], Any]] = ...,
        fset: Optional[Callable[[Any, Any], None]] = ...,
        fdel: Optional[Callable[[Any], None]] = ...,
        doc: Optional[str] = ...,
    ) -> None: ...
    def getter(self, fget: Callable[[Any], Any]) -> property: ...
    def setter(self, fset: Callable[[Any, Any], None]) -> property: ...
    def deleter(self, fdel: Callable[[Any], None]) -> property: ...
    def __get__(self, obj: Any, type: Optional[type] = ...) -> Any: ...
    def __set__(self, obj: Any, value: Any) -> None: ...
    def __delete__(self, obj: Any) -> None: ...
    def fget(self) -> Any: ...
    def fset(self, value: Any) -> None: ...
    def fdel(self) -> None: ...

if sys.version_info < (3,):
    long = int

NotImplemented: Any

def abs(__n: SupportsAbs[_T]) -> _T: ...
def all(__i: Iterable[object]) -> bool: ...
def any(__i: Iterable[object]) -> bool: ...

if sys.version_info < (3,):
    def apply(
        __func: Callable[..., _T],
        __args: Optional[Sequence[Any]] = ...,
        __kwds: Optional[Mapping[str, Any]] = ...,
    ) -> _T: ...

if sys.version_info >= (3,):
    def ascii(__o: object) -> str: ...

def bin(__number: int) -> str: ...

if sys.version_info >= (3, 7):
    def breakpoint(*args: Any, **kws: Any) -> None: ...

def callable(__o: object) -> bool: ...
def chr(__code: int) -> str: ...

if sys.version_info < (3,):
    def cmp(__x: Any, __y: Any) -> int: ...
    _N1 = TypeVar("_N1", bool, int, float, complex)
    def coerce(__x: _N1, __y: _N1) -> Tuple[_N1, _N1]: ...

if sys.version_info >= (3, 6):
    # This class is to be exported as PathLike from os,
    # but we define it here as _PathLike to avoid import cycle issues.
    # See https://github.com/python/typeshed/pull/991#issuecomment-288160993
    class _PathLike(Generic[AnyStr]):
        def __fspath__(self) -> AnyStr: ...
    def compile(
        source: Union[str, bytes, mod, AST],
        filename: Union[str, bytes, _PathLike[Any]],
        mode: str,
        flags: int = ...,
        dont_inherit: int = ...,
        optimize: int = ...,
    ) -> Any: ...

elif sys.version_info >= (3,):
    def compile(
        source: Union[str, bytes, mod, AST],
        filename: Union[str, bytes],
        mode: str,
        flags: int = ...,
        dont_inherit: int = ...,
        optimize: int = ...,
    ) -> Any: ...

else:
    def compile(
        source: Union[Text, mod],
        filename: Text,
        mode: Text,
        flags: int = ...,
        dont_inherit: int = ...,
    ) -> Any: ...

if sys.version_info >= (3,):
    def copyright() -> None: ...
    def credits() -> None: ...

def delattr(__o: Any, __name: Text) -> None: ...
def dir(__o: object = ...) -> List[str]: ...

_N2 = TypeVar("_N2", int, float)

def divmod(__a: _N2, __b: _N2) -> Tuple[_N2, _N2]: ...
def eval(
    __source: Union[Text, bytes, CodeType],
    __globals: Optional[Dict[str, Any]] = ...,
    __locals: Optional[Mapping[str, Any]] = ...,
) -> Any: ...

if sys.version_info >= (3,):
    def exec(
        __object: Union[str, bytes, CodeType],
        __globals: Optional[Dict[str, Any]] = ...,
        __locals: Optional[Mapping[str, Any]] = ...,
    ) -> Any: ...

else:
    def execfile(
        __filename: str,
        __globals: Optional[Dict[str, Any]] = ...,
        __locals: Optional[Dict[str, Any]] = ...,
    ) -> None: ...

def exit(code: object = ...) -> NoReturn: ...

if sys.version_info >= (3,):
    @overload
    def filter(__function: None, __iterable: Iterable[Optional[_T]]) -> Iterator[_T]: ...
    @overload
    def filter(__function: Callable[[_T], Any], __iterable: Iterable[_T]) -> Iterator[_T]: ...

else:
    @overload
    def filter(__function: Callable[[AnyStr], Any], __iterable: AnyStr) -> AnyStr: ...  # type: ignore
    @overload
    def filter(__function: None, __iterable: Tuple[Optional[_T], ...]) -> Tuple[_T, ...]: ...  # type: ignore
    @overload
    def filter(__function: Callable[[_T], Any], __iterable: Tuple[_T, ...]) -> Tuple[_T, ...]: ...  # type: ignore
    @overload
    def filter(__function: None, __iterable: Iterable[Optional[_T]]) -> List[_T]: ...
    @overload
    def filter(__function: Callable[[_T], Any], __iterable: Iterable[_T]) -> List[_T]: ...

def format(__o: object, __format_spec: str = ...) -> str: ...  # TODO unicode
def getattr(__o: Any, name: Text, __default: Any = ...) -> Any: ...
def globals() -> Dict[str, Any]: ...
def hasattr(__o: Any, __name: Text) -> bool: ...
def hash(__o: object) -> int: ...

if sys.version_info >= (3,):
    def help(*args: Any, **kwds: Any) -> None: ...

def hex(__i: Union[int, _SupportsIndex]) -> str: ...
def id(__o: object) -> int: ...

if sys.version_info >= (3,):
    def input(__prompt: Any = ...) -> str: ...

else:
    def input(__prompt: Any = ...) -> Any: ...
    def intern(__string: str) -> str: ...

@overload
def iter(__iterable: Iterable[_T]) -> Iterator[_T]: ...
@overload
def iter(__function: Callable[[], Optional[_T]], __sentinel: None) -> Iterator[_T]: ...
@overload
def iter(__function: Callable[[], _T], __sentinel: Any) -> Iterator[_T]: ...
def isinstance(__o: object, __t: Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]]) -> bool: ...
def issubclass(
    __cls: type, __classinfo: Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]]
) -> bool: ...
def len(__o: Sized) -> int: ...

if sys.version_info >= (3,):
    def license() -> None: ...

def locals() -> Dict[str, Any]: ...

if sys.version_info >= (3,):
    @overload
    def map(__func: Callable[[_T1], _S], __iter1: Iterable[_T1]) -> Iterator[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2], _S], __iter1: Iterable[_T1], __iter2: Iterable[_T2]
    ) -> Iterator[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
    ) -> Iterator[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3, _T4], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
    ) -> Iterator[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3, _T4, _T5], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
        __iter5: Iterable[_T5],
    ) -> Iterator[_S]: ...
    @overload
    def map(
        __func: Callable[..., _S],
        __iter1: Iterable[Any],
        __iter2: Iterable[Any],
        __iter3: Iterable[Any],
        __iter4: Iterable[Any],
        __iter5: Iterable[Any],
        __iter6: Iterable[Any],
        *iterables: Iterable[Any],
    ) -> Iterator[_S]: ...

else:
    @overload
    def map(__func: None, __iter1: Iterable[_T1]) -> List[_T1]: ...
    @overload
    def map(
        __func: None, __iter1: Iterable[_T1], __iter2: Iterable[_T2]
    ) -> List[Tuple[_T1, _T2]]: ...
    @overload
    def map(
        __func: None, __iter1: Iterable[_T1], __iter2: Iterable[_T2], __iter3: Iterable[_T3]
    ) -> List[Tuple[_T1, _T2, _T3]]: ...
    @overload
    def map(
        __func: None,
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
    ) -> List[Tuple[_T1, _T2, _T3, _T4]]: ...
    @overload
    def map(
        __func: None,
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
        __iter5: Iterable[_T5],
    ) -> List[Tuple[_T1, _T2, _T3, _T4, _T5]]: ...
    @overload
    def map(
        __func: None,
        __iter1: Iterable[Any],
        __iter2: Iterable[Any],
        __iter3: Iterable[Any],
        __iter4: Iterable[Any],
        __iter5: Iterable[Any],
        __iter6: Iterable[Any],
        *iterables: Iterable[Any],
    ) -> List[Tuple[Any, ...]]: ...
    @overload
    def map(__func: Callable[[_T1], _S], __iter1: Iterable[_T1]) -> List[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2], _S], __iter1: Iterable[_T1], __iter2: Iterable[_T2]
    ) -> List[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
    ) -> List[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3, _T4], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
    ) -> List[_S]: ...
    @overload
    def map(
        __func: Callable[[_T1, _T2, _T3, _T4, _T5], _S],
        __iter1: Iterable[_T1],
        __iter2: Iterable[_T2],
        __iter3: Iterable[_T3],
        __iter4: Iterable[_T4],
        __iter5: Iterable[_T5],
    ) -> List[_S]: ...
    @overload
    def map(
        __func: Callable[..., _S],
        __iter1: Iterable[Any],
        __iter2: Iterable[Any],
        __iter3: Iterable[Any],
        __iter4: Iterable[Any],
        __iter5: Iterable[Any],
        __iter6: Iterable[Any],
        *iterables: Iterable[Any],
    ) -> List[_S]: ...

if sys.version_info >= (3,):
    @overload
    def max(__arg1: _T, __arg2: _T, *_args: _T, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def max(__iterable: Iterable[_T], *, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def max(
        __iterable: Iterable[_T], *, key: Callable[[_T], Any] = ..., default: _VT
    ) -> Union[_T, _VT]: ...

else:
    @overload
    def max(__arg1: _T, __arg2: _T, *_args: _T, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def max(__iterable: Iterable[_T], *, key: Callable[[_T], Any] = ...) -> _T: ...

if sys.version_info >= (3,):
    @overload
    def min(__arg1: _T, __arg2: _T, *_args: _T, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def min(__iterable: Iterable[_T], *, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def min(
        __iterable: Iterable[_T], *, key: Callable[[_T], Any] = ..., default: _VT
    ) -> Union[_T, _VT]: ...

else:
    @overload
    def min(__arg1: _T, __arg2: _T, *_args: _T, key: Callable[[_T], Any] = ...) -> _T: ...
    @overload
    def min(__iterable: Iterable[_T], *, key: Callable[[_T], Any] = ...) -> _T: ...

@overload
def next(__i: Iterator[_T]) -> _T: ...
@overload
def next(__i: Iterator[_T], default: _VT) -> Union[_T, _VT]: ...
def oct(__i: Union[int, _SupportsIndex]) -> str: ...

if sys.version_info >= (3, 6):
    def open(
        file: Union[str, bytes, int, _PathLike[Any]],
        mode: str = ...,
        buffering: int = ...,
        encoding: Optional[str] = ...,
        errors: Optional[str] = ...,
        newline: Optional[str] = ...,
        closefd: bool = ...,
        opener: Optional[Callable[[str, int], int]] = ...,
    ) -> IO[Any]: ...

elif sys.version_info >= (3,):
    def open(
        file: Union[str, bytes, int],
        mode: str = ...,
        buffering: int = ...,
        encoding: Optional[str] = ...,
        errors: Optional[str] = ...,
        newline: Optional[str] = ...,
        closefd: bool = ...,
        opener: Optional[Callable[[str, int], int]] = ...,
    ) -> IO[Any]: ...

else:
    def open(name: Union[unicode, int], mode: unicode = ..., buffering: int = ...) -> BinaryIO: ...

def ord(__c: Union[Text, bytes]) -> int: ...

if sys.version_info >= (3,):
    class _Writer(Protocol):
        def write(self, __s: str) -> Any: ...
    def print(
        *values: object,
        sep: Text = ...,
        end: Text = ...,
        file: Optional[_Writer] = ...,
        flush: bool = ...,
    ) -> None: ...

else:
    class _Writer(Protocol):
        def write(self, __s: Any) -> Any: ...
    # This is only available after from __future__ import print_function.
    def print(
        *values: object, sep: Text = ..., end: Text = ..., file: Optional[_Writer] = ...
    ) -> None: ...

@overload
def pow(__x: int, __y: int) -> Any: ...  # The return type can be int or float, depending on y
@overload
def pow(__x: int, __y: int, __z: int) -> Any: ...
@overload
def pow(__x: float, __y: float) -> float: ...
@overload
def pow(__x: float, __y: float, __z: float) -> float: ...
def quit(code: object = ...) -> NoReturn: ...

if sys.version_info < (3,):
    def range(__x: int, __y: int = ..., __step: int = ...) -> List[int]: ...
    def raw_input(__prompt: Any = ...) -> str: ...
    @overload
    def reduce(
        __function: Callable[[_T, _S], _T], __iterable: Iterable[_S], __initializer: _T
    ) -> _T: ...
    @overload
    def reduce(__function: Callable[[_T, _T], _T], __iterable: Iterable[_T]) -> _T: ...
    def reload(__module: Any) -> Any: ...

@overload
def reversed(__object: Sequence[_T]) -> Iterator[_T]: ...
@overload
def reversed(__object: Reversible[_T]) -> Iterator[_T]: ...
def repr(__o: object) -> str: ...
@overload
def round(number: float) -> int: ...
@overload
# def round(number: float, ndigits: None) -> int: ...
# @overload
def round(number: float, ndigits: int) -> float: ...
@overload
def round(number: SupportsRound[_T]) -> int: ...
@overload
def round(number: SupportsRound[_T], ndigits: None) -> int: ...
@overload
def round(number: SupportsRound[_T], ndigits: int) -> _T: ...
def setattr(__object: Any, __name: Text, __value: Any) -> None: ...

if sys.version_info >= (3,):
    def sorted(
        __iterable: Iterable[_T], *, key: Optional[Callable[[_T], Any]] = ..., reverse: bool = ...
    ) -> List[_T]: ...

else:
    def sorted(
        __iterable: Iterable[_T],
        *,
        cmp: Callable[[_T, _T], int] = ...,
        key: Optional[Callable[[_T], Any]] = ...,
        reverse: bool = ...,
    ) -> List[_T]: ...

@overload
def sum(__iterable: Iterable[_T]) -> Union[_T, int]: ...
@overload
def sum(__iterable: Iterable[_T], __start: _S) -> Union[_T, _S]: ...

if sys.version_info < (3,):
    def unichr(__i: int) -> unicode: ...

def vars(__object: Any = ...) -> Dict[str, Any]: ...
@overload
def zip(__iter1: Iterable[_T1]) -> Iterator[Tuple[_T1]]: ...
@overload
def zip(__iter1: Iterable[_T1], __iter2: Iterable[_T2]) -> Iterator[Tuple[_T1, _T2]]: ...
@overload
def zip(
    __iter1: Iterable[_T1], __iter2: Iterable[_T2], __iter3: Iterable[_T3]
) -> Iterator[Tuple[_T1, _T2, _T3]]: ...
@overload
def zip(
    __iter1: Iterable[_T1], __iter2: Iterable[_T2], __iter3: Iterable[_T3], __iter4: Iterable[_T4]
) -> Iterator[Tuple[_T1, _T2, _T3, _T4]]: ...
@overload
def zip(
    __iter1: Iterable[_T1],
    __iter2: Iterable[_T2],
    __iter3: Iterable[_T3],
    __iter4: Iterable[_T4],
    __iter5: Iterable[_T5],
) -> Iterator[Tuple[_T1, _T2, _T3, _T4, _T5]]: ...
@overload
def zip(
    __iter1: Iterable[Any],
    __iter2: Iterable[Any],
    __iter3: Iterable[Any],
    __iter4: Iterable[Any],
    __iter5: Iterable[Any],
    __iter6: Iterable[Any],
    *iterables: Iterable[Any],
) -> Iterator[Tuple[Any, ...]]: ...
def __import__(
    name: Text,
    globals: Optional[Mapping[str, Any]] = ...,
    locals: Optional[Mapping[str, Any]] = ...,
    fromlist: Sequence[str] = ...,
    level: int = ...,
) -> Any: ...

# Actually the type of Ellipsis is <type 'ellipsis'>, but since it's
# not exposed anywhere under that name, we make it private here.
class ellipsis: ...

Ellipsis: ellipsis

if sys.version_info < (3,):
    # TODO: buffer support is incomplete; e.g. some_string.startswith(some_buffer) doesn't type check.
    _AnyBuffer = TypeVar("_AnyBuffer", str, unicode, bytearray, buffer)
    class buffer(Sized):
        def __init__(self, object: _AnyBuffer, offset: int = ..., size: int = ...) -> None: ...
        def __add__(self, other: _AnyBuffer) -> str: ...
        def __cmp__(self, other: _AnyBuffer) -> bool: ...
        def __getitem__(self, key: Union[int, slice]) -> str: ...
        def __getslice__(self, i: int, j: int) -> str: ...
        def __len__(self) -> int: ...
        def __mul__(self, x: int) -> str: ...

class BaseException(object):
    args: Tuple[Any, ...]
    if sys.version_info < (3,):
        message: Any
    if sys.version_info >= (3,):
        __cause__: Optional[BaseException]
        __context__: Optional[BaseException]
        __suppress_context__: bool
        __traceback__: Optional[TracebackType]
    def __init__(self, *args: object) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    if sys.version_info < (3,):
        def __getitem__(self, i: int) -> Any: ...
        def __getslice__(self, start: int, stop: int) -> Tuple[Any, ...]: ...
    if sys.version_info >= (3,):
        def with_traceback(self, tb: Optional[TracebackType]) -> BaseException: ...

class GeneratorExit(BaseException): ...
class KeyboardInterrupt(BaseException): ...

class SystemExit(BaseException):
    code: int

class Exception(BaseException): ...

class StopIteration(Exception):
    if sys.version_info >= (3,):
        value: Any

if sys.version_info >= (3,):
    _StandardError = Exception
    class OSError(Exception):
        errno: int
        strerror: str
        # filename, filename2 are actually Union[str, bytes, None]
        filename: Any
        filename2: Any
    EnvironmentError = OSError
    IOError = OSError
else:
    class StandardError(Exception): ...
    _StandardError = StandardError
    class EnvironmentError(StandardError):
        errno: int
        strerror: str
        # TODO can this be unicode?
        filename: str
    class OSError(EnvironmentError): ...
    class IOError(EnvironmentError): ...

class ArithmeticError(_StandardError): ...
class AssertionError(_StandardError): ...
class AttributeError(_StandardError): ...
class BufferError(_StandardError): ...
class EOFError(_StandardError): ...

class ImportError(_StandardError):
    if sys.version_info >= (3,):
        name: str
        path: str

class LookupError(_StandardError): ...
class MemoryError(_StandardError): ...
class NameError(_StandardError): ...
class ReferenceError(_StandardError): ...
class RuntimeError(_StandardError): ...

if sys.version_info >= (3, 5):
    class StopAsyncIteration(Exception):
        value: Any

class SyntaxError(_StandardError):
    msg: str
    lineno: int
    offset: Optional[int]
    text: Optional[str]
    filename: str

class SystemError(_StandardError): ...
class TypeError(_StandardError): ...
class ValueError(_StandardError): ...
class FloatingPointError(ArithmeticError): ...
class OverflowError(ArithmeticError): ...
class ZeroDivisionError(ArithmeticError): ...

if sys.version_info >= (3, 6):
    class ModuleNotFoundError(ImportError): ...

class IndexError(LookupError): ...
class KeyError(LookupError): ...
class UnboundLocalError(NameError): ...

class WindowsError(OSError):
    winerror: int

if sys.version_info >= (3,):
    class BlockingIOError(OSError):
        characters_written: int
    class ChildProcessError(OSError): ...
    class ConnectionError(OSError): ...
    class BrokenPipeError(ConnectionError): ...
    class ConnectionAbortedError(ConnectionError): ...
    class ConnectionRefusedError(ConnectionError): ...
    class ConnectionResetError(ConnectionError): ...
    class FileExistsError(OSError): ...
    class FileNotFoundError(OSError): ...
    class InterruptedError(OSError): ...
    class IsADirectoryError(OSError): ...
    class NotADirectoryError(OSError): ...
    class PermissionError(OSError): ...
    class ProcessLookupError(OSError): ...
    class TimeoutError(OSError): ...

class NotImplementedError(RuntimeError): ...

if sys.version_info >= (3, 5):
    class RecursionError(RuntimeError): ...

class IndentationError(SyntaxError): ...
class TabError(IndentationError): ...
class UnicodeError(ValueError): ...

class UnicodeDecodeError(UnicodeError):
    encoding: str
    object: bytes
    start: int
    end: int
    reason: str
    def __init__(
        self, __encoding: str, __object: bytes, __start: int, __end: int, __reason: str
    ) -> None: ...

class UnicodeEncodeError(UnicodeError):
    encoding: str
    object: Text
    start: int
    end: int
    reason: str
    def __init__(
        self, __encoding: str, __object: Text, __start: int, __end: int, __reason: str
    ) -> None: ...

class UnicodeTranslateError(UnicodeError): ...
class Warning(Exception): ...
class UserWarning(Warning): ...
class DeprecationWarning(Warning): ...
class SyntaxWarning(Warning): ...
class RuntimeWarning(Warning): ...
class FutureWarning(Warning): ...
class PendingDeprecationWarning(Warning): ...
class ImportWarning(Warning): ...
class UnicodeWarning(Warning): ...
class BytesWarning(Warning): ...

if sys.version_info >= (3, 2):
    class ResourceWarning(Warning): ...

if sys.version_info < (3,):
    class file(BinaryIO):
        @overload
        def __init__(self, file: str, mode: str = ..., buffering: int = ...) -> None: ...
        @overload
        def __init__(self, file: unicode, mode: str = ..., buffering: int = ...) -> None: ...
        @overload
        def __init__(self, file: int, mode: str = ..., buffering: int = ...) -> None: ...
        def __iter__(self) -> Iterator[str]: ...
        def next(self) -> str: ...
        def read(self, n: int = ...) -> str: ...
        def __enter__(self) -> BinaryIO: ...
        def __exit__(
            self,
            t: Optional[type] = ...,
            exc: Optional[BaseException] = ...,
            tb: Optional[Any] = ...,
        ) -> Optional[bool]: ...
        def flush(self) -> None: ...
        def fileno(self) -> int: ...
        def isatty(self) -> bool: ...
        def close(self) -> None: ...
        def readable(self) -> bool: ...
        def writable(self) -> bool: ...
        def seekable(self) -> bool: ...
        def seek(self, offset: int, whence: int = ...) -> int: ...
        def tell(self) -> int: ...
        def readline(self, limit: int = ...) -> str: ...
        def readlines(self, hint: int = ...) -> List[str]: ...
        def write(self, data: str) -> int: ...
        def writelines(self, data: Iterable[str]) -> None: ...
        def truncate(self, pos: Optional[int] = ...) -> int: ...



from typing import Tuple

from astroid import nodes

from python_ta.transforms.type_inference_visitor import NoType
from python_ta.typecheck.base import TypeFail, TypeInfo

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type


def generate_tuple(length: int, t: type = None):
    program = ""
    for i in range(length + 1):
        if t is None:
            program += f"x{i}, "
        elif t == int:
            program += f"{i}, "
        elif t == bool:
            program += f"{i % 2 == 0}, "
        elif t == str:
            program += f"'{str(chr(i+97))}', "
    return program


def generate_tuple_assign(length: int, t: type, same_length: bool, more_args: bool = None):
    program = ""
    for i in range(1, length):
        for j in range(1, length):
            if same_length and i == j:
                program += generate_tuple(i) + "= " + generate_tuple(j, t) + "\n"
            elif not same_length:
                if (more_args and i > j) or (not more_args and i < j):
                    program += generate_tuple(i) + "= " + generate_tuple(j, t) + "\n"
    return program


def test_tuple_same_length_int():
    program = generate_tuple_assign(10, int, True)
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert assign_node.inf_type == NoType()


def test_tuple_same_length_bool():
    program = generate_tuple_assign(10, bool, True)
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert assign_node.inf_type == NoType()


def test_tuple_same_length_str():
    program = generate_tuple_assign(10, str, True)
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert assign_node.inf_type == NoType()


def test_tuple_single_var():
    program = """
    a = 1, 2
    b = 1, 2, 3
    c = 1, 2, 3, 4
    """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert assign_node.inf_type == NoType()


def test_tuple_single_val():
    program = """
    a, b = 1
    a, b, c = 1
    a, b, c, d = 1
    """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert isinstance(assign_node.inf_type, TypeFail)


def test_tuple_extra_vars():
    program = generate_tuple_assign(10, int, False, True)
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert isinstance(assign_node.inf_type, TypeFail)


def test_tuple_extra_value():
    program = generate_tuple_assign(10, int, False, False)
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert isinstance(assign_node.inf_type, TypeFail)


def test_tuple_subscript():
    program = """
    lst = ['Hello', 'Goodbye']
    lst[0], lst[1] = 'Bonjour', 'Au revoir'
    """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert not isinstance(assign_node.inf_type, TypeFail)


def test_tuple_attribute():
    program = """
    class A:
        def __init__(self):
            self.first_attr = 0
            self.second_attr = 1

    a = A()
    a.first_attr, a.second_attr = 10, 11
    """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert not isinstance(assign_node.inf_type, TypeFail)


def test_tuple_attribute_variable():
    program = """
    class A:
        def __init__(self):
            self.first_attr = 0
            self.second_attr = 1

    a = A()
    some_list = [1, 2]

    a.first_attr, a.second_attr = some_list
    """
    module, _ = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.Assign):
        assert not isinstance(assign_node.inf_type, TypeFail)


def test_tuple_empty():
    program = """
    def f(x):
        a = ()
        b = (x,)
        a = b
    """
    module, ti = cs._parse_text(program)
    functiondef_node = next(module.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, functiondef_node, "a") == Tuple[()]
    x_type = lookup_type(ti, functiondef_node, "x")
    assert lookup_type(ti, functiondef_node, "b") == Tuple[x_type]



"""Patch pylint checker behaviour.
"""
import astroid
from astroid.bases import BUILTINS
from pylint.checkers.base import NameChecker
from pylint.checkers.classes import ClassChecker
from pylint.checkers.utils import node_frame_class

from python_ta.checkers.global_variables_checker import is_in_main


def patch_checkers():
    """Run patches to modify built-in pylint checker behaviour."""
    _override_check_protected_attribute_access()
    _override_check_invalid_name_in_main()
    _override_attribute_defined_outside_init()


def _override_check_protected_attribute_access():
    """Override protected-member-access check.

    We find pylint's default protected-member-access check too restrictive in
    method bodies; it only allows protected attribute access on the 'self'
    parameter (and from the class itself).

    We change this behaviour to allow access to any protected attribute that is
    defined for this class. (This leads to false negatives unless we combine
    this with type inference, but we're okay with that.)
    """
    old_check_protected_attribute_access = ClassChecker._check_protected_attribute_access

    def _check(self, node):
        attrname = node.attrname
        klass = node_frame_class(node)
        if klass is None or (
            attrname not in klass.instance_attrs
            and attrname not in (m.name for m in klass.methods())
        ):
            old_check_protected_attribute_access(self, node)

    ClassChecker._check_protected_attribute_access = _check


def _override_check_invalid_name_in_main():
    """Override invalid-name check for variables in main block.

    pylint normally complains about variable names in the main block
    that aren't in ALL_CAPS -- in other words, it assumes that all such
    variables should be constants. We disable this check here so that
    non-constant variable names are permitted (encourages experimentation
    in the main block).
    """
    old_visit_assignname = NameChecker.visit_assignname

    def patched_visit_assignname(self, node):
        if is_in_main(node):
            self._check_name("variable", node.name, node)
        else:
            old_visit_assignname(self, node)

    NameChecker.visit_assignname = patched_visit_assignname


def _override_attribute_defined_outside_init():
    """Eliminate attribute-defined-outside-init error when using properties.

    Checks for properties using the "a = property(_get_a, _set_a)" approach
    (no decorator support yet), and allows for a member to be set in the setter
    as long as the setter is called (implicitly) in __init__.
    """
    old_leave_classdef = ClassChecker.leave_classdef

    def new_leave_classdef(self, cnode):
        for attr, nodes in cnode.instance_attrs.items():
            setter = _get_attribute_property_setter(attr, cnode)
            if setter is not None:
                self.config.defining_attr_methods = self.config.defining_attr_methods + (setter,)

        old_leave_classdef(self, cnode)

    ClassChecker.leave_classdef = new_leave_classdef


def _get_attribute_property_setter(name, klass):
    """Return the name of a setter for name in klass, if name is a property."""
    try:
        attributes = klass.getattr(name)
    except astroid.NotFoundError:
        return None

    property_name = "{0}.property".format(BUILTINS)
    for attr in attributes:
        try:
            infered = next(attr.infer())
        except astroid.InferenceError:
            continue
        if infered.pytype() == property_name:
            try:
                # In this case, attr is in a statement of the form
                # <attr> = property(<fget>, <fset>); return the name of <fset>,
                # assuming it is an identifier (and not a more complex expression).
                return attr.parent.value.args[1].name
            except Exception:
                return None



from typing import Any, Dict, List

from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type

settings.load_profile("pyta")


@given(cs.simple_homogeneous_dict_node(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_homogeneous_dict(dictionary):
    """Test Dictionary nodes representing a dictionary with all key:value pairs of same types."""
    module, _ = cs._parse_text(dictionary)
    dict_node = list(module.nodes_of_class(nodes.Dict))[0]
    if len(dict_node.items) == 0:
        assert dict_node.inf_type.getValue() == Dict[Any, Any]
    else:
        first_key, first_value = next(((k, v) for k, v in dictionary.items))
        cs._verify_type_setting(
            module, nodes.Dict, Dict[type(first_key.value), type(first_value.value)]
        )


@given(cs.dict_node(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
def test_heterogeneous_dict(node):
    """Test Dictionary nodes representing a dictionary with some key:value pairs of different types."""
    keys = [item.value for item, _ in node.items]
    values = [item.value for _, item in node.items]
    assume(not isinstance(keys[0], type(keys[1])))
    assume(not isinstance(values[0], type(values[1])))
    assume(not isinstance(keys[1], type(keys[0])))
    assume(not isinstance(values[1], type(values[0])))
    key_types = [type(key.value) for key, _ in node.items]
    val_types = [type(val.value) for _, val in node.items]
    if int in key_types:
        assume(bool not in val_types)
    if bool in key_types:
        assume(int not in val_types)
    module, _ = cs._parse_text(node)
    cs._verify_type_setting(module, nodes.Dict, Dict[Any, Any])


def test_sorted_dict():
    src = """
    dictionary = {'B': 2, 'A': 1}
    sorted_dict = sorted(dictionary)
    """
    module, ti = cs._parse_text(src)
    assign_node = list(module.nodes_of_class(nodes.AssignName))[1]
    t = lookup_type(ti, assign_node, assign_node.name)
    assert ti.type_constraints.resolve(t).getValue() == List[str]


def test_any_dict():
    src = """
    dictionary = {'B': 2, 'A': 1, '': 3}
    any_empty = any(dictionary)
    """
    module, ti = cs._parse_text(src)
    assign_node = list(module.nodes_of_class(nodes.AssignName))[1]
    t = lookup_type(ti, assign_node, assign_node.name)
    assert ti.type_constraints.resolve(t).getValue() == bool
